"""
Manage X.509 certificates
=========================

.. versionadded:: 3006.0

    This module represents a complete rewrite of the original ``x509`` modules
    and is named ``x509_v2`` since it introduces breaking changes.

:depends: cryptography

.. note::

    All parameters that take a public key, private key, certificate,
    CSR or CRL can be specified either as a PEM/hex/base64 string or
    a path to a local file encoded in all supported formats for the type.

Configuration instructions and general remarks are documented
in the :ref:`execution module docs <x509-setup>`.

For the list of breaking changes versus the previous ``x509`` modules,
please also refer to the :ref:`execution module docs <x509-setup>`.

About
-----
This module can enable managing a complete PKI infrastructure, including creating
private keys, CAs, certificates and CRLs. It includes the ability to generate a
private key on a server, and have the corresponding public key sent to a remote
CA to create a CA signed certificate. This can be done in a secure manner, where
private keys are always generated locally and never moved across the network.

Example
-------
Here is a simple example scenario. In this example ``ca`` is the ca server,
and ``www`` is a web server that needs a certificate signed by ``ca``.

.. note::

    Remote signing requires the setup of :term:`Peer Communication` and signing
    policies. Please see the :ref:`execution module docs <x509-setup>`.


/srv/salt/top.sls

.. code-block:: yaml

    base:
      '*':
        - cert
      'ca':
        - ca
      'www':
        - www

This state creates the CA key, certificate and signing policy. It also publishes
the certificate to the mine, where it can be easily retrieved by other minions.

.. code-block:: yaml

    # /srv/salt/ca.sls

    Configure the x509 module:
      file.managed:
        - name: /etc/salt/minion.d/x509.conf
        - source: salt://x509.conf

    Restart Salt minion:
      cmd.run:
        - name: 'salt-call service.restart salt-minion'
        - bg: true
        - onchanges:
          - file: /etc/salt/minion.d/x509.conf

    Ensure PKI directories exist:
      file.directory:
        - name: /etc/pki/issued_certs
        - makedirs: true

    Create CA private key:
      x509.private_key_managed:
        - name: /etc/pki/ca.key
        - keysize: 4096
        - backup: true
        - require:
          - file: /etc/pki/issued_certs

    Create self-signed CA certificate:
      x509.certificate_managed:
        - name: /etc/pki/ca.crt
        - signing_private_key: /etc/pki/ca.key
        - CN: ca.example.com
        - C: US
        - ST: Utah
        - L: Salt Lake City
        - basicConstraints: "critical, CA:true"
        - keyUsage: "critical, cRLSign, keyCertSign"
        - subjectKeyIdentifier: hash
        - authorityKeyIdentifier: keyid:always,issuer
        - days_valid: 3650
        - days_remaining: 0
        - backup: true
        - require:
          - x509: /etc/pki/ca.key

.. code-block:: yaml

    # /srv/salt/x509.conf

    # enable x509_v2
    features:
      x509_v2: true

    # publish the CA certificate to the mine
    mine_functions:
      x509.get_pem_entries: [/etc/pki/ca.crt]

    # define at least one signing policy for remote signing
    x509_signing_policies:
      www:
        - minions: 'www'
        - signing_private_key: /etc/pki/ca.key
        - signing_cert: /etc/pki/ca.crt
        - C: US
        - ST: Utah
        - L: Salt Lake City
        - basicConstraints: "critical CA:false"
        - keyUsage: "critical keyEncipherment"
        - subjectKeyIdentifier: hash
        - authorityKeyIdentifier: keyid:always,issuer
        - days_valid: 30
        - copypath: /etc/pki/issued_certs/


This example state will instruct all minions to trust certificates signed by
our new CA. Mind that this example works for Debian-based OS only.
Also note the Jinja call to encode the string to JSON, which will avoid
YAML issues with newline characters.

.. code-block:: jinja

    # /srv/salt/cert.sls

    Ensure the CA trust bundle exists:
      file.directory:
        - name: /usr/local/share/ca-certificates

    Ensure our self-signed CA certificate is included:
      x509.pem_managed:
        - name: /usr/local/share/ca-certificates/myca.crt
        - text: {{ salt["mine.get"]("ca", "x509.get_pem_entries")["ca"]["/etc/pki/ca.crt"] | json }}

This state creates a private key, then requests a certificate signed by our CA
according to the www policy.

.. code-block:: yaml

    # /srv/salt/www.sls

    Ensure PKI directory exists:
      file.directory:
        - name: /etc/pki

    Create private key for the certificate:
      x509.private_key_managed:
        - name: /etc/pki/www.key
        - keysize: 4096
        - backup: true
        - require:
          - file: /etc/pki

    Request certificate:
      x509.certificate_managed:
        - name: /etc/pki/www.crt
        - ca_server: ca
        - signing_policy: www
        - private_key: /etc/pki/www.key
        - CN: www.example.com
        - days_remaining: 7
        - backup: true
        - require:
          - x509: /etc/pki/www.key
"""

import base64
import copy
import logging
import os.path
from datetime import datetime, timedelta, timezone

import salt.utils.files
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.state import STATE_INTERNAL_KEYWORDS as _STATE_INTERNAL_KEYWORDS

try:
    import cryptography.x509 as cx509
    from cryptography.exceptions import UnsupportedAlgorithm
    from cryptography.hazmat.primitives import hashes

    import salt.utils.x509 as x509util

    HAS_CRYPTOGRAPHY = True
except ImportError:
    HAS_CRYPTOGRAPHY = False


log = logging.getLogger(__name__)

__virtualname__ = "x509"


def __virtual__():
    if not HAS_CRYPTOGRAPHY:
        return (False, "Could not load cryptography")
    if not __opts__["features"].get("x509_v2"):
        return (
            False,
            "x509_v2 needs to be explicitly enabled by setting `x509_v2: true` "
            "in the minion configuration value `features` until Salt 3008 (Argon).",
        )
    return __virtualname__


def certificate_managed(
    name,
    days_remaining=None,
    ca_server=None,
    signing_policy=None,
    encoding="pem",
    append_certs=None,
    copypath=None,
    prepend_cn=False,
    digest="sha256",
    signing_private_key=None,
    signing_private_key_passphrase=None,
    signing_cert=None,
    public_key=None,
    private_key=None,
    private_key_passphrase=None,
    csr=None,
    subject=None,
    serial_number=None,
    not_before=None,
    not_after=None,
    days_valid=None,
    pkcs12_passphrase=None,
    pkcs12_encryption_compat=False,
    pkcs12_friendlyname=None,
    **kwargs,
):
    """
    Ensure an X.509 certificate is present as specified.

    This function accepts the same arguments as :py:func:`x509.create_certificate <salt.modules.x509_v2.create_certificate>`,
    as well as most ones for `:py:func:`file.managed <salt.states.file.managed>`.

    name
        The path the certificate should be present at.

    days_remaining
        The certificate will be recreated once the remaining certificate validity
        period is less than this number of days.
        Defaults to ``90`` (until v3009) or ``7`` (from v3009 onwards).

    ca_server
        Request a remotely signed certificate from ca_server. For this to
        work, a ``signing_policy`` must be specified, and that same policy
        must be configured on the ca_server.  Also, the Salt master must
        permit peers to call the ``x509.sign_remote_certificate`` function.
        See the :ref:`execution module docs <x509-setup>` for details.

    signing_policy
        The name of a configured signing policy. Parameters specified in there
        are hardcoded and cannot be overridden. This is required for remote signing,
        otherwise optional.

    encoding
        Specify the encoding of the resulting certificate. It can be serialized
        as a ``pem`` (or ``pkcs7_pem``) text file or in several binary formats
        (``der``, ``pkcs7_der``, ``pkcs12``). Defaults to ``pem``.

    append_certs
        A list of additional certificates to append to the new one, e.g. to create a CA chain.

        .. note::

            Mind that when ``der`` encoding is in use, appending certificatees is prohibited.

    copypath
        Create a copy of the issued certificate in PEM format in this directory.
        The file will be named ``<serial_number>.crt`` if prepend_cn is false.

    prepend_cn
        When ``copypath`` is set, prepend the common name of the certificate to
        the file name like so: ``<CN>-<serial_number>.crt``. Defaults to false.

    digest
        The hashing algorithm to use for the signature. Valid values are:
        sha1, sha224, sha256, sha384, sha512, sha512_224, sha512_256, sha3_224,
        sha3_256, sha3_384, sha3_512. Defaults to ``sha256``.
        This will be ignored for ``ed25519`` and ``ed448`` key types.

    signing_private_key
        The private key corresponding to the public key in ``signing_cert``. Required.

    signing_private_key_passphrase
        If ``signing_private_key`` is encrypted, the passphrase to decrypt it.

    signing_cert
        The CA certificate to be used for signing the issued certificate.

    public_key
        The public key the certificate should be issued for. Other ways of passing
        the required information are ``private_key`` and ``csr``. If neither are set,
        the public key of the ``signing_private_key`` will be included, i.e.
        a self-signed certificate is generated.

    private_key
        The private key corresponding to the public key the certificate should
        be issued for. This is one way of specifying the public key that will
        be included in the certificate, the other ones being ``public_key`` and ``csr``.

    private_key_passphrase
        If ``private_key`` is specified and encrypted, the passphrase to decrypt it.

    csr
        A certificate signing request to use as a base for generating the certificate.
        The following information will be respected, depending on configuration:

        * public key
        * extensions, if not otherwise specified (arguments, signing_policy)

    subject
        The subject's distinguished name embedded in the certificate. This is one way of
        passing this information (see ``kwargs`` below for the other).
        This argument will be preferred and allows to control the order of RDNs in the DN
        as well as to embed RDNs with multiple attributes.
        This can be specified as a RFC4514-encoded string (``CN=example.com,O=Example Inc,C=US``,
        mind that the rendered order is reversed from what is embedded), a list
        of RDNs encoded as in RFC4514 (``["C=US", "O=Example Inc", "CN=example.com"]``)
        or a dictionary (``{"CN": "example.com", "C": "US", "O": "Example Inc"}``,
        default ordering).
        Multiple name attributes per RDN are concatenated with a ``+``.

        .. note::

            Parsing of RFC4514 strings requires at least cryptography release 37.

    serial_number
        A serial number to be embedded in the certificate. If unspecified, will
        autogenerate one. This should be an integer, either in decimal or
        hexadecimal notation.

    not_before
        Set a specific date the certificate should not be valid before.
        The format should follow ``%Y-%m-%d %H:%M:%S`` and will be interpreted as GMT/UTC.
        Defaults to the time of issuance.

    not_after
        Set a specific date the certificate should not be valid after.
        The format should follow ``%Y-%m-%d %H:%M:%S`` and will be interpreted as GMT/UTC.
        If unspecified, defaults to the current time plus ``days_valid`` days.

    days_valid
        If ``not_after`` is unspecified, the number of days from the time of issuance
        the certificate should be valid for.
        Defaults to ``365`` (until v3009) or ``30`` (from v3009 onwards).

    pkcs12_passphrase
        When encoding a certificate as ``pkcs12``, encrypt it with this passphrase.

        .. note::

            PKCS12 encryption is very weak and `should not be relied on for security <https://cryptography.io/en/stable/hazmat/primitives/asymmetric/serialization/#cryptography.hazmat.primitives.serialization.pkcs12.serialize_key_and_certificates>`_.

    pkcs12_encryption_compat
        OpenSSL 3 and cryptography v37 switched to a much more secure default
        encryption for PKCS12, which might be incompatible with some systems.
        This forces the legacy encryption. Defaults to False.

    pkcs12_friendlyname
        When encoding a certificate as ``pkcs12``, a name for the certificate can be included.

    kwargs
        Embedded X.509v3 extensions and the subject's distinguished name can be
        controlled via supplemental keyword arguments. See
        :py:func:`x509.create_certificate <salt.modules.x509_v2.create_certificate>`
        for an overview.
    """
    # Deprecation checks vs the old x509 module
    if days_valid is None and not_after is None:
        try:
            salt.utils.versions.warn_until(
                "Potassium",
                "The default value for `days_valid` will change to 30. Please adapt your code accordingly.",
            )
            days_valid = 365
        except RuntimeError:
            days_valid = 30

    if days_remaining is None:
        try:
            salt.utils.versions.warn_until(
                "Potassium",
                "The default value for `days_remaining` will change to 7. Please adapt your code accordingly.",
            )
            days_remaining = 90
        except RuntimeError:
            days_remaining = 7

    if "algorithm" in kwargs:
        salt.utils.versions.warn_until(
            "Potassium",
            "`algorithm` has been renamed to `digest`. Please update your code.",
        )
        digest = kwargs.pop("algorithm")
    kwargs = x509util.ensure_cert_kwargs_compat(kwargs)

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "The certificate is in the correct state",
    }
    current = current_encoding = None
    changes = {}
    verb = "create"
    file_args, cert_args = _split_file_kwargs(_filter_state_internal_kwargs(kwargs))
    append_certs = append_certs or []
    if not isinstance(append_certs, list):
        append_certs = [append_certs]

    try:
        # check file.managed changes early to avoid using unnecessary resources
        file_managed_test = _file_managed(name, test=True, replace=False, **file_args)
        if file_managed_test["result"] is False:
            ret["result"] = False
            ret["comment"] = (
                "Problem while testing file.managed changes, see its output"
            )
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if "is not present and is not set for creation" in file_managed_test["comment"]:
            _add_sub_state_run(ret, file_managed_test)
            return ret

        real_name = name
        replace = False

        # handle follow_symlinks
        if __salt__["file.is_link"](name):
            if file_args.get("follow_symlinks", True):
                real_name = os.path.realpath(name)
            else:
                # workaround https://github.com/saltstack/salt/issues/31802
                __salt__["file.remove"](name)
                replace = True

        if __salt__["file.file_exists"](real_name):
            try:
                (
                    current,
                    current_encoding,
                    current_chain,
                    current_extra,
                ) = x509util.load_cert(
                    real_name, passphrase=pkcs12_passphrase, get_encoding=True
                )
            except SaltInvocationError as err:
                if "Bad decrypt" in str(err):
                    changes["pkcs12_passphrase"] = True
                elif any(
                    (
                        "Could not deserialize binary data" in str(err),
                        "Could not load PEM-encoded" in str(err),
                    )
                ):
                    replace = True
                else:
                    raise
            else:
                if encoding != current_encoding:
                    changes["encoding"] = encoding
                elif encoding == "pkcs12" and current_extra.cert.friendly_name != (
                    salt.utils.stringutils.to_bytes(pkcs12_friendlyname)
                    if pkcs12_friendlyname
                    else None
                ):
                    changes["pkcs12_friendlyname"] = pkcs12_friendlyname
                try:
                    curr_not_after = current.not_valid_after_utc
                except AttributeError:
                    # naive datetime object, release <42 (it's always UTC)
                    curr_not_after = current.not_valid_after.replace(
                        tzinfo=timezone.utc
                    )

                if curr_not_after < datetime.now(tz=timezone.utc) + timedelta(
                    days=days_remaining
                ):
                    changes["expiration"] = True

                current_chain = current_chain or []
                ca_chain = [x509util.load_cert(x) for x in append_certs]
                if not _compare_ca_chain(current_chain, ca_chain):
                    changes["additional_certs"] = True

                (
                    builder,
                    private_key_loaded,
                    signing_cert_loaded,
                    final_kwargs,
                ) = _build_cert(
                    ca_server=ca_server,
                    signing_policy=signing_policy,
                    digest=digest,  # passed because of signing_policy merging
                    signing_private_key=signing_private_key,
                    signing_private_key_passphrase=signing_private_key_passphrase,
                    signing_cert=signing_cert,
                    public_key=public_key,
                    private_key=private_key,
                    private_key_passphrase=private_key_passphrase,
                    csr=csr,
                    subject=subject,
                    serial_number=serial_number,
                    not_before=not_before,
                    not_after=not_after,
                    days_valid=days_valid,
                    **cert_args,
                )

                try:
                    if current.signature_hash_algorithm is not None and not isinstance(
                        current.signature_hash_algorithm,
                        type(x509util.get_hashing_algorithm(final_kwargs["digest"])),
                    ):
                        # ed25519, ed448 do not use a separate hash for signatures, hence algo is None
                        changes["digest"] = digest
                except UnsupportedAlgorithm:
                    # this eg happens with sha3 in cryptography < v39
                    log.warning(
                        "Could not determine signature hash algorithm of '%s'. "
                        "Continuing anyways",
                        name,
                    )

                changes.update(
                    _compare_cert(
                        current,
                        builder,
                        signing_cert=signing_cert_loaded,
                        serial_number=serial_number,
                        not_before=not_before,
                        not_after=not_after,
                    )
                )
        else:
            changes["created"] = name

        if replace:
            changes["replaced"] = name

        if (
            not changes
            and file_managed_test["result"]
            and not file_managed_test["changes"]
        ):
            _add_sub_state_run(ret, file_managed_test)
            return ret

        ret["changes"] = changes
        if current and changes:
            verb = "recreate"

        if __opts__["test"]:
            ret["result"] = None if changes else True
            ret["comment"] = (
                f"The certificate would have been {verb}d"
                if changes
                else ret["comment"]
            )
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if changes:
            if not set(changes) - {
                "additional_certs",
                "encoding",
                "pkcs12_friendlyname",
            }:
                # do not reissue if only metaparameters changed
                if encoding == "pkcs12":
                    cert = __salt__["x509.encode_certificate"](
                        current,
                        append_certs=append_certs,
                        encoding=encoding,
                        private_key=private_key_loaded,
                        pkcs12_passphrase=pkcs12_passphrase,
                        pkcs12_encryption_compat=pkcs12_encryption_compat,
                        pkcs12_friendlyname=pkcs12_friendlyname,
                    )
                else:
                    cert = __salt__["x509.encode_certificate"](
                        current, encoding=encoding, append_certs=append_certs
                    )
            else:
                # request a new certificate otherwise
                cert = __salt__["x509.create_certificate"](
                    ca_server=ca_server,
                    signing_policy=signing_policy,
                    encoding=encoding,
                    append_certs=append_certs,
                    pkcs12_passphrase=pkcs12_passphrase,
                    pkcs12_encryption_compat=pkcs12_encryption_compat,
                    pkcs12_friendlyname=pkcs12_friendlyname,
                    digest=digest,
                    signing_private_key=signing_private_key,
                    signing_private_key_passphrase=signing_private_key_passphrase,
                    signing_cert=signing_cert,
                    public_key=public_key,
                    private_key=private_key,
                    private_key_passphrase=private_key_passphrase,
                    csr=csr,
                    subject=subject,
                    serial_number=serial_number,
                    not_before=not_before,
                    not_after=not_after,
                    days_valid=days_valid,
                    **cert_args,
                )
            ret["comment"] = f"The certificate has been {verb}d"
            if encoding not in ["pem", "pkcs7_pem"]:
                # file.managed does not support binary contents, so create
                # an empty file first (makedirs). This will not work with check_cmd!
                file_managed_ret = _file_managed(name, replace=False, **file_args)
                _add_sub_state_run(ret, file_managed_ret)
                if not _check_file_ret(file_managed_ret, ret, current):
                    return ret
                _safe_atomic_write(
                    real_name, base64.b64decode(cert), file_args.get("backup", "")
                )

        if not changes or encoding in ["pem", "pkcs7_pem"]:
            replace = bool(encoding in ["pem", "pkcs7_pem"] and changes)
            contents = cert if replace else None
            file_managed_ret = _file_managed(
                name, contents=contents, replace=replace, **file_args
            )
            _add_sub_state_run(ret, file_managed_ret)
            if not _check_file_ret(file_managed_ret, ret, current):
                return ret

    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}
    return ret


def crl_managed(
    name,
    signing_private_key,
    revoked,
    days_remaining=None,
    signing_cert=None,
    signing_private_key_passphrase=None,
    include_expired=False,
    days_valid=None,
    digest="sha256",
    encoding="pem",
    extensions=None,
    **kwargs,
):
    """
    Ensure a certificate revocation list is present as specified.

    This function accepts the same arguments as :py:func:`x509.create_crl <salt.modules.x509_v2.create_crl>`,
    as well as most ones for `:py:func:`file.managed <salt.states.file.managed>`.

    name
        The path the certificate revocation list should be present at.

    signing_private_key
        Your certificate authority's private key. It will be used to sign
        the CRL. Required.

    revoked
        A list of dicts containing all the certificates to revoke. Each dict
        represents one certificate. A dict must contain either the key
        ``serial_number`` with the value of the serial number to revoke, or
        ``certificate`` with some reference to the certificate to revoke.

        The dict can optionally contain the ``revocation_date`` key. If this
        key is omitted, the revocation date will be set to now. It should be a
        string in the format ``%Y-%m-%d %H:%M:%S``.

        The dict can also optionally contain the ``not_after`` key. This is
        redundant if the ``certificate`` key is included. If the
        ``certificate`` key is not included, this can be used for the logic
        behind the ``include_expired`` parameter. It should be a string in
        the format ``%Y-%m-%d %H:%M:%S``.

        The dict can also optionally contain the ``extensions`` key, which
        allows to set CRL entry-specific extensions. The following extensions
        are supported:

        certificateIssuer
            Identifies the certificate issuer associated with an entry in an
            indirect CRL. The format is the same as for subjectAltName.

        CRLReason
            Identifies the reason for certificate revocation.
            Available choices are ``unspecified``, ``keyCompromise``, ``CACompromise``,
            ``affiliationChanged``, ``superseded``, ``cessationOfOperation``,
            ``certificateHold``, ``privilegeWithdrawn``, ``aACompromise`` and
            ``removeFromCRL``.

        invalidityDate
            Provides the date on which the certificate became invalid.
            The value should be a string in the same format as ``revocation_date``.

    days_remaining
        The certificate revocation list will be recreated once the remaining
        CRL validity period is less than this number of days.
        Defaults to ``30`` (until v3009) or ``3`` (from v3009 onwards).
        Set to 0 to disable automatic renewal without anything changing.

    signing_cert
        The CA certificate to be used for signing the issued certificate.

    signing_private_key_passphrase
        If ``signing_private_key`` is encrypted, the passphrase to decrypt it.

    include_expired
        Also include already expired certificates in the CRL. Defaults to false.

    days_valid
        The number of days that the CRL should be valid for. This sets the ``Next Update``
        field in the CRL. Defaults to ``100`` (until v3009) or ``7`` (from v3009 onwards).

    digest
        The hashing algorithm to use for the signature. Valid values are:
        sha1, sha224, sha256, sha384, sha512, sha512_224, sha512_256, sha3_224,
        sha3_256, sha3_384, sha3_512. Defaults to ``sha256``.
        This will be ignored for ``ed25519`` and ``ed448`` key types.

    encoding
        Specify the encoding of the resulting certificate revocation list.
        It can be serialized as a ``pem`` text or binary ``der`` file.
        Defaults to ``pem``.

    extensions
        Add CRL extensions. See :py:func:`x509.create_crl <salt.modules.x509_v2.create_crl>`
        for details.

        .. note::

            For ``cRLNumber``, in addition the value ``auto`` is supported, which
            automatically increases the counter every time a new CRL is issued.

    Example:

    .. code-block:: yaml

        Manage CRL:
          x509.crl_managed:
            - name: /etc/pki/ca.crl
            - signing_private_key: /etc/pki/myca.key
            - signing_cert: /etc/pki/myca.crt
            - revoked:
              - certificate: /etc/pki/certs/badweb.crt
                revocation_date: 2022-11-01 00:00:00
                extensions:
                  CRLReason: keyCompromise
              - serial_number: D6:D2:DC:D8:4D:5C:C0:F4
                not_after: 2023-03-14 00:00:00
                revocation_date: 2022-10-25 00:00:00
                extensions:
                  CRLReason: cessationOfOperation
            - extensions:
                cRLNumber: auto
    """
    if "text" in kwargs:
        salt.utils.versions.kwargs_warn_until(["text"], "Potassium")
        kwargs.pop("text")

    if days_valid is None:
        try:
            salt.utils.versions.warn_until(
                "Potassium",
                "The default value for `days_valid` will change to 7. Please adapt your code accordingly.",
            )
            days_valid = 100
        except RuntimeError:
            days_valid = 7

    if days_remaining is None:
        try:
            salt.utils.versions.warn_until(
                "Potassium",
                "The default value for `days_remaining` will change to 3. Please adapt your code accordingly.",
            )
            days_remaining = 30
        except RuntimeError:
            days_remaining = 3

    revoked_parsed = []
    for rev in revoked:
        parsed = {}
        if len(rev) == 1 and isinstance(rev[next(iter(rev))], list):
            salt.utils.versions.warn_until(
                "Potassium",
                "Revoked certificates should be specified as a simple list of dicts.",
            )
            for val in rev[next(iter(rev))]:
                parsed.update(val)
        if "reason" in (parsed or rev):
            salt.utils.versions.warn_until(
                "Potassium",
                "The `reason` parameter for revoked certificates should be specified in extensions:CRLReason.",
            )
            salt.utils.dictupdate.set_dict_key_value(
                (parsed or rev), "extensions:CRLReason", (parsed or rev).pop("reason")
            )
        revoked_parsed.append(parsed or rev)
    revoked = revoked_parsed

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "The certificate revocation list is in the correct state",
    }
    current = current_encoding = None
    changes = {}
    verb = "create"
    file_args, extra_args = _split_file_kwargs(_filter_state_internal_kwargs(kwargs))
    extensions = extensions or {}
    if extra_args:
        raise SaltInvocationError(f"Unrecognized keyword arguments: {list(extra_args)}")

    try:
        # check file.managed changes early to avoid using unnecessary resources
        file_managed_test = _file_managed(name, test=True, replace=False, **file_args)

        if file_managed_test["result"] is False:
            ret["result"] = False
            ret["comment"] = (
                "Problem while testing file.managed changes, see its output"
            )
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if "is not present and is not set for creation" in file_managed_test["comment"]:
            _add_sub_state_run(ret, file_managed_test)
            return ret

        real_name = name
        replace = False

        # handle follow_symlinks
        if __salt__["file.is_link"](name):
            if file_args.get("follow_symlinks", True):
                real_name = os.path.realpath(name)
            else:
                # workaround https://github.com/saltstack/salt/issues/31802
                __salt__["file.remove"](name)
                replace = True

        if __salt__["file.file_exists"](real_name):
            try:
                current, current_encoding = x509util.load_crl(
                    real_name, get_encoding=True
                )
            except SaltInvocationError as err:
                if any(
                    (
                        "Could not load PEM-encoded" in str(err),
                        "Could not load DER-encoded" in str(err),
                    )
                ):
                    replace = True
                else:
                    raise
            else:
                try:
                    if current.signature_hash_algorithm is not None and not isinstance(
                        current.signature_hash_algorithm,
                        type(x509util.get_hashing_algorithm(digest)),
                    ):
                        # ed25519, ed448 do not use a separate hash for signatures, hence algo is None
                        # although CA certificates should not be using those currently
                        changes["digest"] = digest
                except UnsupportedAlgorithm:
                    # this eg happens with sha3 digest in cryptography < v39
                    log.warning(
                        "Could not determine signature hash algorithm of '%s'. "
                        "Continuing anyways",
                        name,
                    )

                if encoding != current_encoding:
                    changes["encoding"] = encoding
                try:
                    curr_next_update = current.next_update_utc
                except AttributeError:
                    # naive datetime object, release <42 (it's always UTC)
                    curr_next_update = current.next_update.replace(tzinfo=timezone.utc)
                if days_remaining and (
                    curr_next_update
                    < datetime.now(tz=timezone.utc) + timedelta(days=days_remaining)
                ):
                    changes["expiration"] = True

                # "auto" is a value that is managed in this function and cannot not be compared
                crl_auto = extensions.get("cRLNumber") == "auto"
                if crl_auto:
                    extensions.pop("cRLNumber")

                builder, sig_privkey = x509util.build_crl(
                    signing_private_key,
                    revoked,
                    signing_cert=signing_cert,
                    signing_private_key_passphrase=signing_private_key_passphrase,
                    include_expired=include_expired,
                    days_valid=days_valid,
                    extensions=extensions,
                )
                changes.update(_compare_crl(current, builder, sig_privkey.public_key()))
                if crl_auto:
                    # put cRLNumber = auto back if it was set
                    extensions["cRLNumber"] = "auto"
                    changes["extensions"]["removed"].pop(
                        changes["extensions"]["removed"].index("cRLNumber")
                    )
                    if not any(changes["extensions"].values()):
                        changes.pop("extensions")
        else:
            changes["created"] = name

        if replace:
            changes["replaced"] = name

        if (
            not changes
            and file_managed_test["result"]
            and not file_managed_test["changes"]
        ):
            _add_sub_state_run(ret, file_managed_test)
            return ret

        ret["changes"] = changes
        if current and changes:
            verb = "recreate"

        if __opts__["test"]:
            ret["result"] = None if changes else True
            ret["comment"] = (
                f"The certificate revocation list would have been {verb}d"
                if changes
                else ret["comment"]
            )
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if changes:
            if not set(changes) - {"encoding"}:
                # do not regenerate if only metaparameters changed
                crl = __salt__["x509.encode_crl"](current, encoding=encoding)
            else:
                # autoincrease cRLNumber counter, if requested
                if extensions.get("cRLNumber") == "auto":
                    try:
                        extensions["cRLNumber"] = (
                            current.extensions.get_extension_for_class(
                                cx509.CRLNumber
                            ).value.crl_number
                            + 1
                        )
                    except (AttributeError, cx509.ExtensionNotFound):
                        extensions["cRLNumber"] = 1
                crl = __salt__["x509.create_crl"](
                    signing_private_key,
                    revoked,
                    signing_cert=signing_cert,
                    signing_private_key_passphrase=signing_private_key_passphrase,
                    include_expired=include_expired,
                    days_valid=days_valid,
                    digest=digest,
                    encoding=encoding,
                    extensions=extensions,
                )
            ret["comment"] = f"The certificate revocation list has been {verb}d"
            if encoding == "der":
                # file.managed does not support binary contents, so create
                # an empty file first (makedirs). This will not work with check_cmd!
                file_managed_ret = _file_managed(name, replace=False, **file_args)
                _add_sub_state_run(ret, file_managed_ret)
                if not _check_file_ret(file_managed_ret, ret, current):
                    return ret
                _safe_atomic_write(
                    real_name, base64.b64decode(crl), file_args.get("backup", "")
                )

        if not changes or encoding == "pem":
            replace = bool((encoding == "pem") and changes)
            contents = crl if replace else None
            file_managed_ret = _file_managed(
                name, contents=contents, replace=replace, **file_args
            )
            _add_sub_state_run(ret, file_managed_ret)
            if not _check_file_ret(file_managed_ret, ret, current):
                return ret
    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}
    return ret


def csr_managed(
    name,
    private_key,
    private_key_passphrase=None,
    digest="sha256",
    encoding="pem",
    subject=None,
    **kwargs,
):
    """
    Ensure a certificate signing request is present as specified.

    This function accepts the same arguments as :py:func:`x509.create_csr <salt.modules.x509_v2.create_csr>`,
    as well as most ones for :py:func:`file.managed <salt.states.file.managed>`.

    name
        The path the certificate signing request should be present at.

    private_key
        The private key corresponding to the public key the certificate should
        be issued for. The CSR will be signed by it. Required.

    private_key_passphrase
        If ``private_key`` is encrypted, the passphrase to decrypt it.

    digest
        The hashing algorithm to use for the signature. Valid values are:
        sha1, sha224, sha256, sha384, sha512, sha512_224, sha512_256, sha3_224,
        sha3_256, sha3_384, sha3_512. Defaults to ``sha256``.
        This will be ignored for ``ed25519`` and ``ed448`` key types.

    encoding
        Specify the encoding of the resulting certificate revocation list.
        It can be serialized as a ``pem`` text or binary ``der`` file.
        Defaults to ``pem``.

    kwargs
        Embedded X.509v3 extensions and the subject's distinguished name can be
        controlled via supplemental keyword arguments.
        See :py:func:`x509.create_certificate <salt.modules.x509_v2.create_certificate>`
        for an overview. Mind that some extensions are not available for CSR
        (``authorityInfoAccess``, ``authorityKeyIdentifier``,
        ``issuerAltName``, ``crlDistributionPoints``).
    """
    # Deprecation checks vs the old x509 module
    if "algorithm" in kwargs:
        salt.utils.versions.warn_until(
            "Potassium",
            "`algorithm` has been renamed to `digest`. Please update your code.",
        )
        digest = kwargs.pop("algorithm")
    kwargs = x509util.ensure_cert_kwargs_compat(kwargs)

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "The certificate signing request is in the correct state",
    }
    current = current_encoding = None
    changes = {}
    verb = "create"
    file_args, csr_args = _split_file_kwargs(_filter_state_internal_kwargs(kwargs))

    try:
        # check file.managed changes early to avoid using unnecessary resources
        file_managed_test = _file_managed(name, test=True, replace=False, **file_args)

        if file_managed_test["result"] is False:
            ret["result"] = False
            ret["comment"] = (
                "Problem while testing file.managed changes, see its output"
            )
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if "is not present and is not set for creation" in file_managed_test["comment"]:
            _add_sub_state_run(ret, file_managed_test)
            return ret

        real_name = name
        replace = False

        # handle follow_symlinks
        if __salt__["file.is_link"](name):
            if file_args.get("follow_symlinks", True):
                real_name = os.path.realpath(name)
            else:
                # workaround https://github.com/saltstack/salt/issues/31802
                __salt__["file.remove"](name)
                replace = True

        if __salt__["file.file_exists"](real_name):
            try:
                current, current_encoding = x509util.load_csr(
                    real_name, get_encoding=True
                )
            except SaltInvocationError as err:
                if any(
                    (
                        "Could not load PEM-encoded" in str(err),
                        "Could not load DER-encoded" in str(err),
                    )
                ):
                    replace = True
                else:
                    raise
            except cx509.InvalidVersion:
                # by default, the previous x509 modules generated CSR with
                # invalid versions, which leads to an exception in cryptography >= v38
                changes["invalid_version"] = True
                replace = True
            else:
                try:
                    if current.signature_hash_algorithm is not None and not isinstance(
                        current.signature_hash_algorithm,
                        type(x509util.get_hashing_algorithm(digest)),
                    ):
                        # ed25519, ed448 do not use a separate hash for signatures, hence algo is None
                        changes["digest"] = digest
                except UnsupportedAlgorithm:
                    # this eg happens with sha3 digest in cryptography < v39
                    log.warning(
                        "Could not determine signature hash algorithm of '%s'. "
                        "Continuing anyways",
                        name,
                    )

                if encoding != current_encoding:
                    changes["encoding"] = encoding

                builder, privkey = x509util.build_csr(
                    private_key,
                    private_key_passphrase=private_key_passphrase,
                    subject=subject,
                    **csr_args,
                )
                if not x509util.is_pair(current.public_key(), privkey):
                    changes["private_key"] = True
                changes.update(_compare_csr(current, builder))
        else:
            changes["created"] = name

        if replace:
            changes["replaced"] = name

        if (
            not changes
            and file_managed_test["result"]
            and not file_managed_test["changes"]
        ):
            _add_sub_state_run(ret, file_managed_test)
            return ret

        ret["changes"] = changes
        if current and changes:
            verb = "recreate"

        if __opts__["test"]:
            ret["result"] = None if changes else True
            ret["comment"] = (
                f"The certificate signing request would have been {verb}d"
                if changes
                else ret["comment"]
            )
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if changes:
            if not set(changes) - {"encoding"}:
                # do not regenerate if only metaparameters changed
                csr = __salt__["x509.encode_csr"](current, encoding=encoding)
            else:
                csr = __salt__["x509.create_csr"](
                    private_key,
                    private_key_passphrase=private_key_passphrase,
                    digest=digest,
                    encoding=encoding,
                    subject=subject,
                    **csr_args,
                )
            ret["comment"] = f"The certificate signing request has been {verb}d"
            if encoding == "der":
                # file.managed does not support binary contents, so create
                # an empty file first (makedirs). This will not work with check_cmd!
                file_managed_ret = _file_managed(name, replace=False, **file_args)
                _add_sub_state_run(ret, file_managed_ret)
                if not _check_file_ret(file_managed_ret, ret, current):
                    return ret
                _safe_atomic_write(
                    real_name, base64.b64decode(csr), file_args.get("backup", "")
                )
        if not changes or encoding == "pem":
            replace = bool((encoding == "pem") and changes)
            contents = csr if replace else None
            file_managed_ret = _file_managed(
                name, contents=contents, replace=replace, **file_args
            )
            _add_sub_state_run(ret, file_managed_ret)
            if not _check_file_ret(file_managed_ret, ret, current):
                return ret

    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}
    return ret


def pem_managed(name, text, **kwargs):
    """
    Manage the contents of a PEM file directly with the content in text,
    ensuring correct formatting.

    name
        The path to the file to manage.

    text
        The PEM-formatted text to write.

    kwargs
        Most arguments supported by :py:func:`file.managed <salt.states.file.managed>` are passed through.
    """
    file_args, extra_args = _split_file_kwargs(kwargs)
    if extra_args:
        raise SaltInvocationError(f"Unrecognized keyword arguments: {list(extra_args)}")

    try:
        file_args["contents"] = __salt__["x509.get_pem_entry"](text=text)
    except (CommandExecutionError, SaltInvocationError) as err:
        return {"name": name, "result": False, "comment": str(err), "changes": {}}
    return _file_managed(name, **file_args)


def private_key_managed(
    name,
    algo="rsa",
    keysize=None,
    passphrase=None,
    encoding="pem",
    new=False,
    overwrite=False,
    pkcs12_encryption_compat=False,
    **kwargs,
):
    """
    Ensure a private key is present as specified.

    This function accepts the same arguments as :py:func:`x509.create_private_key <salt.modules.x509_v2.create_private_key>`,
    as well as most ones for :py:func:`file.managed <salt.states.file.managed>`.

    .. note::

        If ``mode`` is unspecified, it will default to ``0400``.

    name
        The path the private key should be present at.

    algo
        The digital signature scheme the private key should be based on.
        Available: ``rsa``, ``ec``, ``ed25519``, ``ed448``. Defaults to ``rsa``.

    keysize
        For ``rsa``, specifies the bitlength of the private key (2048, 3072, 4096).
        For ``ec``, specifies the NIST curve to use (256, 384, 521).
        Irrelevant for Edwards-curve schemes (``ed25519``, ``ed448``).
        Defaults to 2048 for RSA and 256 for EC.

    passphrase
        If this is specified, the private key will be encrypted using this
        passphrase. The encryption algorithm cannot be selected, it will be
        determined automatically as the best available one.

    encoding
        Specify the encoding of the resulting private key. It can be serialized
        as a ``pem`` text, binary ``der`` or ``pkcs12`` file.
        Defaults to ``pem``.

    new
        Always create a new key. Defaults to false.
        Combining new with :mod:`prereq <salt.states.requisites.prereq>`
        can allow key rotation whenever a new certificate is generated.

    overwrite
        Overwrite an existing private key if the provided passphrase cannot decrypt it.
        Defaults to false.

    pkcs12_encryption_compat
        Some operating systems are incompatible with the encryption defaults
        for PKCS12 used since OpenSSL v3. This switch triggers a fallback to
        ``PBESv1SHA1And3KeyTripleDESCBC``.
        Please consider the `notes on PKCS12 encryption <https://cryptography.io/en/stable/hazmat/primitives/asymmetric/serialization/#cryptography.hazmat.primitives.serialization.pkcs12.serialize_key_and_certificates>`_.

    Example:

    The Jinja templating in this example ensures a new private key is generated
    if the file does not exist and whenever the associated certificate
    is to be renewed.

    .. code-block:: jinja

        Manage www private key:
          x509.private_key_managed:
            - name: /etc/pki/www.key
            - keysize: 4096
            - new: true
        {%- if salt["file.file_exists"]("/etc/pki/www.key") %}
            - prereq:
              - x509: /etc/pki/www.crt
        {%- endif %}
    """
    # Deprecation checks vs the old x509 module
    if "bits" in kwargs:
        salt.utils.versions.warn_until(
            "Potassium",
            "`bits` has been renamed to `keysize`. Please update your code.",
        )
        keysize = kwargs.pop("bits")

    ignored_params = {"cipher", "verbose", "text"}.intersection(
        kwargs
    )  # path, overwrite
    if ignored_params:
        salt.utils.versions.kwargs_warn_until(ignored_params, "Potassium")
        for x in ignored_params:
            kwargs.pop(x)

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "The private key is in the correct state",
    }
    current = current_encoding = None
    changes = {}
    verb = "create"
    file_args, extra_args = _split_file_kwargs(kwargs)

    if extra_args:
        raise SaltInvocationError(f"Unrecognized keyword arguments: {list(extra_args)}")

    if not file_args.get("mode"):
        # ensure secure defaults
        file_args["mode"] = "0400"

    try:
        if keysize and algo in ["ed25519", "ed448"]:
            raise SaltInvocationError(f"keysize is an invalid parameter for {algo}")

        # check file.managed changes early to avoid using unnecessary resources
        file_managed_test = _file_managed(name, test=True, replace=False, **file_args)

        if file_managed_test["result"] is False:
            ret["result"] = False
            ret["comment"] = (
                "Problem while testing file.managed changes, see its output"
            )
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if "is not present and is not set for creation" in file_managed_test["comment"]:
            _add_sub_state_run(ret, file_managed_test)
            return ret

        real_name = name
        replace = False

        # handle follow_symlinks
        if __salt__["file.is_link"](name):
            if file_args.get("follow_symlinks", True):
                real_name = os.path.realpath(name)
            else:
                # workaround https://github.com/saltstack/salt/issues/31802
                __salt__["file.remove"](name)
                replace = True

        file_exists = __salt__["file.file_exists"](real_name)

        if file_exists and not new:
            try:
                current, current_encoding, _ = x509util.load_privkey(
                    real_name, passphrase=passphrase, get_encoding=True
                )
            except SaltInvocationError as err:
                if "Bad decrypt" in str(err):
                    if not overwrite:
                        raise CommandExecutionError(
                            "The provided passphrase cannot decrypt the private key. "
                            "Pass overwrite: true to force regeneration"
                        ) from err
                    changes["passphrase"] = True
                elif any(
                    (
                        "Could not deserialize binary data" in str(err),
                        "Could not load DER-encoded" in str(err),
                        "Could not load PEM-encoded" in str(err),
                    )
                ):
                    if not overwrite:
                        raise CommandExecutionError(
                            "The existing file does not seem to be a private key "
                            "formatted as DER, PEM or embedded in PKCS12. "
                            "Pass overwrite: true to force regeneration"
                        ) from err
                    replace = True
                elif "Private key is unencrypted" in str(err):
                    changes["passphrase"] = True
                    current, current_encoding, _ = x509util.load_privkey(
                        real_name, passphrase=None, get_encoding=True
                    )
                elif "Private key is encrypted" in str(err) and not passphrase:
                    if not overwrite:
                        raise CommandExecutionError(
                            "The existing file is encrypted. Pass overwrite: true "
                            "to force regeneration without passphrase"
                        ) from err
                    changes["passphrase"] = True
                else:
                    raise
        if current:
            key_type = x509util.get_key_type(current)
            check_keysize = keysize
            if check_keysize is None:
                if algo == "rsa":
                    check_keysize = 2048
                elif algo == "ec":
                    check_keysize = 256
            if any(
                (
                    (algo == "rsa" and not key_type == x509util.KEY_TYPE.RSA),
                    (algo == "ec" and not key_type == x509util.KEY_TYPE.EC),
                    (algo == "ed25519" and not key_type == x509util.KEY_TYPE.ED25519),
                    (algo == "ed448" and not key_type == x509util.KEY_TYPE.ED448),
                )
            ):
                changes["algo"] = algo
            if (
                "algo" not in changes
                and algo in ("rsa", "ec")
                and current.key_size != check_keysize
            ):
                changes["keysize"] = check_keysize
            if encoding != current_encoding:
                changes["encoding"] = encoding
        elif file_exists and new:
            changes["replaced"] = name
        else:
            changes["created"] = name

        if (
            not changes
            and file_managed_test["result"]
            and not file_managed_test["changes"]
        ):
            _add_sub_state_run(ret, file_managed_test)
            return ret

        ret["changes"] = changes
        if file_exists and changes:
            verb = "recreate"

        if __opts__["test"]:
            ret["result"] = None if changes else True
            ret["comment"] = (
                f"The private key would have been {verb}d"
                if changes
                else ret["comment"]
            )
            _add_sub_state_run(ret, file_managed_test)
            return ret

        if changes:
            if not set(changes) - {"encoding", "passphrase"}:
                # do not regenerate if only metaparameters changed
                pk = __salt__["x509.encode_private_key"](
                    current, passphrase=passphrase, encoding=encoding
                )
            else:
                pk = __salt__["x509.create_private_key"](
                    algo=algo,
                    keysize=keysize,
                    passphrase=passphrase,
                    encoding=encoding,
                    pkcs12_encryption_compat=pkcs12_encryption_compat,
                )
            ret["comment"] = f"The private key has been {verb}d"
            if encoding != "pem":
                # file.managed does not support binary contents, so create
                # an empty file first (makedirs). This will not work with check_cmd!
                file_managed_ret = _file_managed(name, replace=False, **file_args)
                _add_sub_state_run(ret, file_managed_ret)
                if not _check_file_ret(file_managed_ret, ret, current):
                    return ret
                _safe_atomic_write(
                    real_name, base64.b64decode(pk), file_args.get("backup", "")
                )

        if not changes or encoding == "pem":
            replace = bool((encoding == "pem") and changes)
            contents = pk if replace else None
            file_managed_ret = _file_managed(
                name, contents=contents, replace=replace, **file_args
            )
            _add_sub_state_run(ret, file_managed_ret)
            if not _check_file_ret(file_managed_ret, ret, current):
                return ret
    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}
    return ret


def _filter_state_internal_kwargs(kwargs):
    # check_cmd is a valid argument to file.managed
    ignore = set(_STATE_INTERNAL_KEYWORDS) - {"check_cmd"}
    return {k: v for k, v in kwargs.items() if k not in ignore}


def _split_file_kwargs(kwargs):
    valid_file_args = [
        "user",
        "group",
        "mode",
        "attrs",
        "makedirs",
        "dir_mode",
        "backup",
        "create",
        "follow_symlinks",
        "check_cmd",
        "tmp_dir",
        "tmp_ext",
        "selinux",
        "encoding",
        "encoding_errors",
        "win_owner",
        "win_perms",
        "win_deny_perms",
        "win_inheritance",
        "win_perms_reset",
    ]
    file_args = {"show_changes": False}
    extra_args = {}
    for k, v in kwargs.items():
        if k in valid_file_args:
            file_args[k] = v
        else:
            extra_args[k] = v
    return file_args, extra_args


def _add_sub_state_run(ret, sub):
    sub["low"] = {
        "name": ret["name"],
        "state": "file",
        "__id__": __low__["__id__"],
        "fun": "managed",
    }
    if "sub_state_run" not in ret:
        ret["sub_state_run"] = []
    ret["sub_state_run"].append(sub)


def _file_managed(name, test=None, **kwargs):
    if test not in [None, True]:
        raise SaltInvocationError("test param can only be None or True")
    # work around https://github.com/saltstack/salt/issues/62590
    test = test or __opts__["test"]
    res = __salt__["state.single"]("file.managed", name, test=test, **kwargs)
    return res[next(iter(res))]


def _check_file_ret(fret, ret, current):
    if fret["result"] is False:
        ret["result"] = False
        ret["comment"] = (
            f"Could not {'create' if not current else 'update'} file, see file.managed output"
        )
        ret["changes"] = {}
        return False
    return True


def _build_cert(
    ca_server=None, signing_policy=None, signing_private_key=None, **kwargs
):
    final_kwargs = copy.deepcopy(kwargs)
    final_kwargs["signing_private_key"] = signing_private_key
    x509util.merge_signing_policy(
        __salt__["x509.get_signing_policy"](signing_policy, ca_server=ca_server),
        final_kwargs,
    )
    signing_private_key = final_kwargs.pop("signing_private_key")

    builder, _, private_key_loaded, signing_cert = x509util.build_crt(
        signing_private_key,
        skip_load_signing_private_key=ca_server is not None,
        **final_kwargs,
    )
    return builder, private_key_loaded, signing_cert, final_kwargs


def _compare_cert(current, builder, signing_cert, serial_number, not_before, not_after):
    changes = {}

    if (
        serial_number is not None
        and _getattr_safe(builder, "_serial_number") != current.serial_number
    ):
        changes["serial_number"] = serial_number

    if not x509util.match_pubkey(
        _getattr_safe(builder, "_public_key"), current.public_key()
    ):
        changes["private_key"] = True

    if signing_cert and not x509util.verify_signature(
        current, signing_cert.public_key()
    ):
        changes["signing_private_key"] = True

    if _getattr_safe(builder, "_subject_name") != current.subject:
        changes["subject_name"] = _getattr_safe(
            builder, "_subject_name"
        ).rfc4514_string()

    if _getattr_safe(builder, "_issuer_name") != current.issuer:
        changes["issuer_name"] = _getattr_safe(builder, "_issuer_name").rfc4514_string()

    ext_changes = _compare_exts(current, builder)
    if any(ext_changes.values()):
        changes["extensions"] = ext_changes
    return changes


def _compare_csr(current, builder):
    changes = {}

    # if _getattr_safe(builder, "_subject_name") != current.subject:
    if not _compareattr_safe(builder, "_subject_name", current.subject):
        changes["subject_name"] = _getattr_safe(
            builder, "_subject_name"
        ).rfc4514_string()

    ext_changes = _compare_exts(current, builder)
    if any(ext_changes.values()):
        changes["extensions"] = ext_changes
    return changes


def _compare_crl(current, builder, sig_pubkey):
    # these are necessary because the classes do not have the required method
    def _get_revoked_certificate_by_serial_number(revoked, serial):
        try:
            return [x for x in revoked if x.serial_number == serial][0]
        except IndexError:
            return None

    def _get_extension_for_oid(extensions, oid):
        try:
            return [x for x in extensions if x.oid == oid][0]
        except IndexError:
            return None

    changes = {}

    if _getattr_safe(builder, "_issuer_name") != current.issuer:
        changes["issuer_name"] = _getattr_safe(builder, "_issuer_name").rfc4514_string()
    if not current.is_signature_valid(sig_pubkey):
        changes["public_key"] = True

    rev_changes = {"added": [], "changed": [], "removed": []}
    revoked = _getattr_safe(builder, "_revoked_certificates")
    for rev in revoked:
        cur = current.get_revoked_certificate_by_serial_number(rev.serial_number)
        if cur is None:
            # certificate was not revoked before
            rev_changes["added"].append(x509util.dec2hex(rev.serial_number))
            continue

        for ext in rev.extensions:
            cur_ext = _get_extension_for_oid(cur.extensions, ext.oid)
            # revoked certificate's extensions have changed (added/changed)
            if any(
                (
                    cur_ext is None,
                    cur_ext.critical != ext.critical,
                    cur_ext.value != ext.value,
                )
            ):
                rev_changes["changed"].append(x509util.dec2hex(rev.serial_number))

        for cur_ext in cur.extensions:
            if _get_extension_for_oid(rev.extensions, cur_ext.oid) is None:
                # an extension was removed from from the revoked certificate
                rev_changes["changed"].append(x509util.dec2hex(rev.serial_number))

    for rev in current:
        # certificate was removed from the CRL, probably because it was outdated anyways
        if (
            _get_revoked_certificate_by_serial_number(revoked, rev.serial_number)
            is None
        ):
            rev_changes["removed"].append(x509util.dec2hex(rev.serial_number))

    if any(rev_changes.values()):
        changes["revocations"] = rev_changes

    ext_changes = _compare_exts(current, builder)
    if any(ext_changes.values()):
        changes["extensions"] = ext_changes
    return changes


def _compare_exts(current, builder):
    def getextname(ext):
        try:
            return ext.oid._name
        except AttributeError:
            return ext.oid.dotted_string

    added = []
    changed = []
    removed = []
    builder_extensions = cx509.Extensions(_getattr_safe(builder, "_extensions"))

    # iter is unnecessary, but avoids a pylint < 2.13.6 crash
    for ext in iter(builder_extensions):
        try:
            cur_ext = current.extensions.get_extension_for_oid(ext.value.oid)
            if cur_ext.critical != ext.critical or cur_ext.value != ext.value:
                changed.append(getextname(ext))
        except cx509.ExtensionNotFound:
            added.append(getextname(ext))

    for ext in current.extensions:
        try:
            builder_extensions.get_extension_for_oid(ext.value.oid)
        except cx509.ExtensionNotFound:
            removed.append(getextname(ext))

    return {"added": added, "changed": changed, "removed": removed}


def _compare_ca_chain(current, new):
    if not len(current) == len(new):
        return False
    for i, new_cert in enumerate(new):
        if new_cert.fingerprint(hashes.SHA256()) != current[i].fingerprint(
            hashes.SHA256()
        ):
            return False
    return True


def _getattr_safe(obj, attr):
    try:
        return getattr(obj, attr)
    except AttributeError as err:
        # Since we cannot get the certificate object without signing,
        # we need to compare attributes marked as internal. At least
        # convert possible exceptions into some description.
        raise CommandExecutionError(
            f"Could not get attribute {attr} from {obj.__class__.__name__}. "
            "Did the internal API of cryptography change?"
        ) from err


def _compareattr_safe(obj, attr, comp):
    try:
        return getattr(obj, attr) == comp
    except AttributeError:
        return False


def _safe_atomic_write(dst, data, backup):
    """
    Create a temporary file with only user r/w perms and atomically
    copy it to the destination, honoring ``backup``.
    """
    tmp = salt.utils.files.mkstemp(prefix=salt.utils.files.TEMPFILE_PREFIX)
    with salt.utils.files.fopen(tmp, "wb") as tmp_:
        tmp_.write(data)
    salt.utils.files.copyfile(
        tmp, dst, __salt__["config.backup_mode"](backup), __opts__["cachedir"]
    )
    salt.utils.files.safe_rm(tmp)



"""
Integration tests for functions located in the salt.cloud.__init__.py file.
"""

import pytest

import salt.cloud
from tests.integration.cloud.helpers.cloud_test_base import CloudTest
from tests.support.helpers import PRE_PYTEST_SKIP


@PRE_PYTEST_SKIP
class CloudClientTestCase(CloudTest):
    """
    Integration tests for the CloudClient class. Uses DigitalOcean as a salt-cloud provider.
    """

    PROVIDER = "digitalocean"
    REQUIRED_PROVIDER_CONFIG_ITEMS = tuple()
    IMAGE_NAME = "14.04.5 x64"

    @pytest.mark.expensive_test
    def setUp(self):

        # Use a --list-images salt-cloud call to see if the DigitalOcean provider is
        # configured correctly before running any tests.
        images = self.run_cloud(f"--list-images {self.PROVIDER}")

        if self.image_name not in [i.strip() for i in images]:
            self.skipTest(
                "Image '{}' was not found in image search. Is the {} provider "
                "configured correctly for this test?".format(
                    self.PROVIDER, self.image_name
                )
            )

    def test_cloud_client_create_and_delete(self):
        """
        Tests that a VM is created successfully when calling salt.cloud.CloudClient.create(),
        which does not require a profile configuration.

        Also checks that salt.cloud.CloudClient.destroy() works correctly since this test needs
        to remove the VM after creating it.

        This test was created as a regression check against Issue #41971.
        """
        cloud_client = salt.cloud.CloudClient(self.config_file)

        # Create the VM using salt.cloud.CloudClient.create() instead of calling salt-cloud
        ret_val = cloud_client.create(
            provider=self.PROVIDER,
            names=[self.instance_name],
            image=self.IMAGE_NAME,
            location="sfo1",
            size="512mb",
            vm_size="512mb",
        )

        # Check that the VM was created correctly
        self.assertInstanceExists(ret_val)

        # Clean up after ourselves and delete the VM
        deleted = cloud_client.destroy(names=[self.instance_name])

        # Check that the VM was deleted correctly
        self.assertIn(self.instance_name, deleted)



import pytest

import salt.modules.nginx as nginx
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {nginx: {}}


@pytest.mark.parametrize(
    "expected_version,nginx_output",
    [
        ("1.2.3", "nginx version: nginx/1.2.3"),
        ("1", "nginx version: nginx/1"),
        ("9.1.100a1+abc123", "nginx version: nginx/9.1.100a1+abc123"),
        (
            "42.9.13.1111111111.whatever",
            "nginx version: nginx/42.9.13.1111111111.whatever",
        ),
    ],
)
def test_basic_nginx_version_output(expected_version, nginx_output):
    with patch.dict(nginx.__salt__, {"cmd.run": lambda *args, **kwargs: nginx_output}):
        assert nginx.version() == expected_version



import logging
import socket
import textwrap

import pytest

import salt.exceptions
import salt.utils.network
import salt.utils.network as network
import salt.utils.platform
from salt._compat import ipaddress
from tests.support.mock import MagicMock, create_autospec, mock_open, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
]


log = logging.getLogger(__name__)

LINUX = """\
eth0      Link encap:Ethernet  HWaddr e0:3f:49:85:6a:af
          inet addr:10.10.10.56  Bcast:10.10.10.255  Mask:255.255.252.0
          inet6 addr: fe80::e23f:49ff:fe85:6aaf/64 Scope:Link
          UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
          RX packets:643363 errors:0 dropped:0 overruns:0 frame:0
          TX packets:196539 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:1000
          RX bytes:386388355 (368.4 MiB)  TX bytes:25600939 (24.4 MiB)

lo        Link encap:Local Loopback
          inet addr:127.0.0.1  Mask:255.0.0.0
          inet6 addr: ::1/128 Scope:Host
          UP LOOPBACK RUNNING  MTU:65536  Metric:1
          RX packets:548901 errors:0 dropped:0 overruns:0 frame:0
          TX packets:548901 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:0
          RX bytes:613479895 (585.0 MiB)  TX bytes:613479895 (585.0 MiB)
"""

FREEBSD = """
em0: flags=8843<UP,BROADCAST,RUNNING,SIMPLEX,MULTICAST> metric 0 mtu 1500
        options=4219b<RXCSUM,TXCSUM,VLAN_MTU,VLAN_HWTAGGING,VLAN_HWCSUM,TSO4,WOL_MAGIC,VLAN_HWTSO>
        ether 00:30:48:ff:ff:ff
        inet 10.10.10.250 netmask 0xffffffe0 broadcast 10.10.10.255
        inet 10.10.10.56 netmask 0xffffffc0 broadcast 10.10.10.63
        media: Ethernet autoselect (1000baseT <full-duplex>)
        status: active
em1: flags=8c02<BROADCAST,OACTIVE,SIMPLEX,MULTICAST> metric 0 mtu 1500
        options=4219b<RXCSUM,TXCSUM,VLAN_MTU,VLAN_HWTAGGING,VLAN_HWCSUM,TSO4,WOL_MAGIC,VLAN_HWTSO>
        ether 00:30:48:aa:aa:aa
        media: Ethernet autoselect
        status: no carrier
plip0: flags=8810<POINTOPOINT,SIMPLEX,MULTICAST> metric 0 mtu 1500
lo0: flags=8049<UP,LOOPBACK,RUNNING,MULTICAST> metric 0 mtu 16384
        options=3<RXCSUM,TXCSUM>
        inet6 fe80::1%lo0 prefixlen 64 scopeid 0x8
        inet6 ::1 prefixlen 128
        inet 127.0.0.1 netmask 0xff000000
        nd6 options=3<PERFORMNUD,ACCEPT_RTADV>
tun0: flags=8051<UP,POINTOPOINT,RUNNING,MULTICAST> metric 0 mtu 1500
        options=80000<LINKSTATE>
        inet 10.12.0.1 --> 10.12.0.2 netmask 0xffffffff
        Opened by PID 1964
"""

SOLARIS = """\
lo0: flags=2001000849<UP,LOOPBACK,RUNNING,MULTICAST,IPv4,VIRTUAL> mtu 8232 index 1
        inet 127.0.0.1 netmask ff000000
net0: flags=100001100943<UP,BROADCAST,RUNNING,PROMISC,MULTICAST,ROUTER,IPv4,PHYSRUNNING> mtu 1500 index 2
        inet 10.10.10.38 netmask ffffffe0 broadcast 10.10.10.63
ilbint0: flags=110001100843<UP,BROADCAST,RUNNING,MULTICAST,ROUTER,IPv4,VRRP,PHYSRUNNING> mtu 1500 index 3
        inet 10.6.0.11 netmask ffffff00 broadcast 10.6.0.255
ilbext0: flags=110001100843<UP,BROADCAST,RUNNING,MULTICAST,ROUTER,IPv4,VRRP,PHYSRUNNING> mtu 1500 index 4
        inet 10.10.11.11 netmask ffffffe0 broadcast 10.10.11.31
ilbext0:1: flags=110001100843<UP,BROADCAST,RUNNING,MULTICAST,ROUTER,IPv4,VRRP,PHYSRUNNING> mtu 1500 index 4
        inet 10.10.11.12 netmask ffffffe0 broadcast 10.10.11.31
vpn0: flags=1000011008d1<UP,POINTOPOINT,RUNNING,NOARP,MULTICAST,ROUTER,IPv4,PHYSRUNNING> mtu 1480 index 5
        inet tunnel src 10.10.11.12 tunnel dst 10.10.5.5
        tunnel hop limit 64
        inet 10.6.0.14 --> 10.6.0.15 netmask ff000000
lo0: flags=2002000849<UP,LOOPBACK,RUNNING,MULTICAST,IPv6,VIRTUAL> mtu 8252 index 1
        inet6 ::1/128
net0: flags=120002004941<UP,RUNNING,PROMISC,MULTICAST,DHCP,IPv6,PHYSRUNNING> mtu 1500 index 2
        inet6 fe80::221:9bff:fefd:2a22/10
ilbint0: flags=120002000840<RUNNING,MULTICAST,IPv6,PHYSRUNNING> mtu 1500 index 3
        inet6 ::/0
ilbext0: flags=120002000840<RUNNING,MULTICAST,IPv6,PHYSRUNNING> mtu 1500 index 4
        inet6 ::/0
vpn0: flags=120002200850<POINTOPOINT,RUNNING,MULTICAST,NONUD,IPv6,PHYSRUNNING> mtu 1480 index 5
        inet tunnel src 10.10.11.12 tunnel dst 10.10.5.5
        tunnel hop limit 64
        inet6 ::/0 --> fe80::b2d6:7c10
"""

NETBSD = """\
vioif0: flags=0x8943<UP,BROADCAST,RUNNING,PROMISC,SIMPLEX,MULTICAST> mtu 1500
        ec_capabilities=1<VLAN_MTU>
        ec_enabled=0
        address: 00:a0:98:e6:83:18
        inet 192.168.1.80/24 broadcast 192.168.1.255 flags 0x0
        inet6 fe80::2a0:98ff:fee6:8318%vioif0/64 flags 0x0 scopeid 0x1
lo0: flags=0x8049<UP,LOOPBACK,RUNNING,MULTICAST> mtu 33624
        inet 127.0.0.1/8 flags 0x0
        inet6 ::1/128 flags 0x20<NODAD>
        inet6 fe80::1%lo0/64 flags 0x0 scopeid 0x2
"""

FREEBSD_SOCKSTAT = """\
USER    COMMAND     PID     FD  PROTO  LOCAL ADDRESS    FOREIGN ADDRESS
root    python2.7   1294    41  tcp4   127.0.0.1:61115  127.0.0.1:4506
"""

FREEBSD_SOCKSTAT_WITH_FAT_PID = """\
USER     COMMAND    PID   FD PROTO  LOCAL ADDRESS    FOREIGN ADDRESS
salt-master python2.781106 35 tcp4  127.0.0.1:61115  127.0.0.1:4506
"""

OPENBSD_NETSTAT = """\
Active Internet connections
Proto   Recv-Q Send-Q  Local Address          Foreign Address        (state)
tcp          0      0  127.0.0.1.61115        127.0.0.1.4506         ESTABLISHED
"""

LINUX_NETLINK_SS_OUTPUT = """\
State       Recv-Q Send-Q                                                            Local Address:Port                                                                           Peer Address:Port
TIME-WAIT   0      0                                                                         [::1]:8009                                                                                  [::1]:40368
LISTEN      0      128                                                                   127.0.0.1:5903                                                                                0.0.0.0:*
ESTAB       0      0                                                            [::ffff:127.0.0.1]:4506                                                                    [::ffff:127.0.0.1]:32315
ESTAB       0      0                                                                 192.168.122.1:4506                                                                       192.168.122.177:24545
ESTAB       0      0                                                                    127.0.0.1:56726                                                                             127.0.0.1:4505
ESTAB       0      0                                                                ::ffff:1.2.3.4:5678                                                                        ::ffff:1.2.3.4:4505
"""

IPV4_SUBNETS = {
    True: ("10.10.0.0/24",),
    False: ("10.10.0.0", "10.10.0.0/33", "FOO", 9, "0.9.800.1000/24"),
}
IPV6_SUBNETS = {
    True: ("::1/128",),
    False: ("::1", "::1/129", "FOO", 9, "aj01::feac/64"),
}


_ip = ipaddress.ip_address


@pytest.fixture(scope="module")
def linux_interfaces_dict():
    return {
        "eth0": {
            "hwaddr": "e0:3f:49:85:6a:af",
            "inet": [
                {
                    "address": "10.10.10.56",
                    "broadcast": "10.10.10.255",
                    "netmask": "255.255.252.0",
                }
            ],
            "inet6": [
                {
                    "address": "fe80::e23f:49ff:fe85:6aaf",
                    "prefixlen": "64",
                    "scope": "link",
                }
            ],
            "up": True,
        },
        "lo": {
            "inet": [{"address": "127.0.0.1", "netmask": "255.0.0.0"}],
            "inet6": [{"address": "::1", "prefixlen": "128", "scope": "host"}],
            "up": True,
        },
    }


@pytest.fixture(scope="module")
def freebsd_interfaces_dict():
    return {
        "": {"up": False},
        "em0": {
            "hwaddr": "00:30:48:ff:ff:ff",
            "inet": [
                {
                    "address": "10.10.10.250",
                    "broadcast": "10.10.10.255",
                    "netmask": "255.255.255.224",
                },
                {
                    "address": "10.10.10.56",
                    "broadcast": "10.10.10.63",
                    "netmask": "255.255.255.192",
                },
            ],
            "up": True,
        },
        "em1": {"hwaddr": "00:30:48:aa:aa:aa", "up": False},
        "lo0": {
            "inet": [{"address": "127.0.0.1", "netmask": "255.0.0.0"}],
            "inet6": [
                {"address": "fe80::1", "prefixlen": "64", "scope": "0x8"},
                {"address": "::1", "prefixlen": "128", "scope": None},
            ],
            "up": True,
        },
        "plip0": {"up": False},
        "tun0": {
            "inet": [{"address": "10.12.0.1", "netmask": "255.255.255.255"}],
            "up": True,
        },
    }


def test_sanitize_host_ip():
    ret = network.sanitize_host("10.1./2.$3")
    assert ret == "10.1.2.3"


def test_sanitize_host_name():
    """
    Should not remove the underscore
    """
    ret = network.sanitize_host("foo_bar")
    assert ret == "foo_bar"


def test_host_to_ips():
    """
    NOTE: When this test fails it's usually because the IP address has
    changed. In these cases, we just need to update the IP address in the
    assertion.
    """

    def getaddrinfo_side_effect(host, *args):
        if host == "github.com":
            return [
                (2, 1, 6, "", ("192.30.255.112", 0)),
                (2, 1, 6, "", ("192.30.255.113", 0)),
            ]
        if host == "ipv6host.foo":
            return [
                (socket.AF_INET6, 1, 6, "", ("2001:a71::1", 0, 0, 0)),
            ]
        raise socket.gaierror(-2, "Name or service not known")

    getaddrinfo_mock = MagicMock(side_effect=getaddrinfo_side_effect)
    with patch.object(socket, "getaddrinfo", getaddrinfo_mock):
        # Test host that can be resolved
        ret = network.host_to_ips("github.com")
        assert ret == ["192.30.255.112", "192.30.255.113"]

        # Test ipv6
        ret = network.host_to_ips("ipv6host.foo")
        assert ret == ["2001:a71::1"]
        # Test host that can't be resolved
        ret = network.host_to_ips("someothersite.com")
        assert ret is None


def test_generate_minion_id():
    assert network.generate_minion_id()


def test__generate_minion_id_with_unicode_in_etc_hosts():
    """
    Test that unicode in /etc/hosts doesn't raise an error when
    _generate_minion_id() helper is called to gather the hosts.
    """
    content = textwrap.dedent(
        """\
    # 以下为主机名解析
    ## ccc
    127.0.0.1       localhost thisismyhostname     # 本机
    """
    )
    fopen_mock = mock_open(read_data={"/etc/hosts": content})
    with patch("salt.utils.files.fopen", fopen_mock):
        assert "thisismyhostname" in network._generate_minion_id()


@pytest.mark.parametrize(
    "addr,expected",
    (
        ("10.10.0.3", True),
        ("0.9.800.1000", False),
        # Check 16-char-long unicode string
        # https://github.com/saltstack/salt/issues/51258
        ("sixteen-char-str", False),
    ),
)
def test_is_ip(addr, expected):
    assert network.is_ip(addr) is expected


@pytest.mark.parametrize(
    "addr,expected",
    (
        ("10.10.0.3", True),
        ("10.100.1", False),
        ("2001:db8:0:1:1:1:1:1", False),
        # Check 16-char-long unicode string
        # https://github.com/saltstack/salt/issues/51258
        ("sixteen-char-str", False),
    ),
)
def test_is_ipv4(addr, expected):
    assert network.is_ipv4(addr) is expected


@pytest.mark.parametrize(
    "addr,expected",
    (
        ("2001:db8:0:1:1:1:1:1", True),
        ("0:0:0:0:0:0:0:1", True),
        ("::1", True),
        ("::", True),
        ("2001:0db8:85a3:0000:0000:8a2e:0370:7334", True),
        ("2001:0db8:85a3::8a2e:0370:7334", True),
        ("2001:0db8:0370:7334", False),
        ("2001:0db8:::0370:7334", False),
        ("10.0.1.2", False),
        ("2001.0db8.85a3.0000.0000.8a2e.0370.7334", False),
        # Check 16-char-long unicode string
        # https://github.com/saltstack/salt/issues/51258
        ("sixteen-char-str", False),
    ),
)
def test_is_ipv6(addr, expected):
    assert network.is_ipv6(addr) is expected


@pytest.mark.parametrize(
    "addr,expected",
    (
        ("2001:db8:0:1:1:1:1:1", "2001:db8:0:1:1:1:1:1"),
        ("0:0:0:0:0:0:0:1", "::1"),
        ("::1", "::1"),
        ("::", "::"),
        ("2001:0db8:85a3:0000:0000:8a2e:0370:7334", "2001:db8:85a3::8a2e:370:7334"),
        ("2001:0db8:85a3::8a2e:0370:7334", "2001:db8:85a3::8a2e:370:7334"),
        ("2001:67c:2e8::/48", "2001:67c:2e8::/48"),
    ),
)
def test_ipv6(addr, expected):
    assert network.ipv6(addr) == expected


@pytest.mark.parametrize(
    "addr,expected",
    (
        ("127.0.1.1", True),
        ("::1", True),
        ("10.0.1.2", False),
        ("2001:db8:0:1:1:1:1:1", False),
    ),
)
def test_is_loopback(addr, expected):
    assert network.is_loopback(addr) is expected


@pytest.mark.parametrize(
    "addr,expected",
    (
        ("10.10.0.3", (_ip("10.10.0.3").compressed, None)),
        ("10.10.0.3:1234", (_ip("10.10.0.3").compressed, 1234)),
        (
            "2001:0db8:85a3::8a2e:0370:7334",
            (
                _ip("2001:0db8:85a3::8a2e:0370:7334").compressed,
                None,
            ),
        ),
        (
            "[2001:0db8:85a3::8a2e:0370:7334]:1234",
            (
                _ip("2001:0db8:85a3::8a2e:0370:7334").compressed,
                1234,
            ),
        ),
        ("2001:0db8:85a3::7334", (_ip("2001:0db8:85a3::7334").compressed, None)),
        (
            "[2001:0db8:85a3::7334]:1234",
            (
                _ip("2001:0db8:85a3::7334").compressed,
                1234,
            ),
        ),
    ),
)
def test_parse_host_port_good(addr, expected):
    assert network.parse_host_port(addr) == expected


@pytest.mark.parametrize(
    "addr",
    (
        "10.10.0.3/24",
        "10.10.0.3::1234",
        "2001:0db8:0370:7334",
        "2001:0db8:0370::7334]:1234",
        "2001:0db8:0370:0:a:b:c:d:1234",
        "host name",
        "host name:1234",
        "10.10.0.3:abcd",
    ),
)
def test_parse_host_port_bad_raises_value_error(addr):
    with pytest.raises(ValueError):
        network.parse_host_port(addr)


@pytest.mark.parametrize(
    "host",
    (
        (
            {
                "host": "10.10.0.3",
                "port": "",
                "mocked": [(2, 1, 6, "", ("10.10.0.3", 0))],
                "ret": "10.10.0.3",
            }
        ),
        (
            {
                "host": "10.10.0.3",
                "port": "1234",
                "mocked": [(2, 1, 6, "", ("10.10.0.3", 0))],
                "ret": "10.10.0.3",
            }
        ),
        (
            {
                "host": "2001:0db8:85a3::8a2e:0370:7334",
                "port": "",
                "mocked": [(10, 1, 6, "", ("2001:db8:85a3::8a2e:370:7334", 0, 0, 0))],
                "ret": "[2001:db8:85a3::8a2e:370:7334]",
            }
        ),
        (
            {
                "host": "2001:0db8:85a3::8a2e:370:7334",
                "port": "1234",
                "mocked": [(10, 1, 6, "", ("2001:db8:85a3::8a2e:370:7334", 0, 0, 0))],
                "ret": "[2001:db8:85a3::8a2e:370:7334]",
            }
        ),
        (
            {
                "host": "salt-master",
                "port": "1234",
                "mocked": [(2, 1, 6, "", ("127.0.0.1", 0))],
                "ret": "127.0.0.1",
            }
        ),
    ),
)
def test_dns_check(host):
    with patch.object(
        socket,
        "getaddrinfo",
        create_autospec(socket.getaddrinfo, return_value=host["mocked"]),
    ):
        with patch("socket.socket", create_autospec(socket.socket)):
            ret = network.dns_check(host["host"], host["port"])
            assert ret == host["ret"]


def test_dns_check_ipv6_filter():
    # raise exception to skip everything after the getaddrinfo call
    with patch.object(
        socket,
        "getaddrinfo",
        create_autospec(socket.getaddrinfo, side_effect=Exception),
    ) as getaddrinfo:
        for ipv6, param in [
            (None, socket.AF_UNSPEC),
            (True, socket.AF_INET6),
            (False, socket.AF_INET),
        ]:
            with pytest.raises(Exception):
                network.dns_check("foo", "1", ipv6=ipv6)
            getaddrinfo.assert_called_with("foo", "1", param, socket.SOCK_STREAM)


def test_dns_check_errors():
    with patch.object(
        socket, "getaddrinfo", create_autospec(socket.getaddrinfo, return_value=[])
    ):
        with pytest.raises(
            salt.exceptions.SaltSystemExit,
            match="DNS lookup or connection check of 'foo' failed.",
        ) as exc_info:
            network.dns_check("foo", "1")

    with patch.object(
        socket,
        "getaddrinfo",
        create_autospec(socket.getaddrinfo, side_effect=TypeError),
    ):
        with pytest.raises(
            salt.exceptions.SaltSystemExit, match="Invalid or unresolveable address"
        ) as exc_info2:
            network.dns_check("foo", "1")


def test_test_addrs():
    # subset of real data from getaddrinfo against saltstack.com
    addrinfo = [
        (30, 2, 17, "", ("2600:9000:21eb:a800:8:1031:abc0:93a1", 0, 0, 0)),
        (30, 1, 6, "", ("2600:9000:21eb:a800:8:1031:abc0:93a1", 0, 0, 0)),
        (30, 2, 17, "", ("2600:9000:21eb:b400:8:1031:abc0:93a1", 0, 0, 0)),
        (30, 1, 6, "", ("2600:9000:21eb:b400:8:1031:abc0:93a1", 0, 0, 0)),
        (2, 1, 6, "", ("13.35.99.52", 0)),
        (2, 2, 17, "", ("13.35.99.85", 0)),
        (2, 1, 6, "", ("13.35.99.85", 0)),
        (2, 2, 17, "", ("13.35.99.122", 0)),
    ]
    with patch("socket.socket", create_autospec(socket.socket)) as s:
        # we connect to the first address
        addrs = network._test_addrs(addrinfo, 80)
        assert len(addrs) == 1
        assert addrs[0] == addrinfo[0][4][0]

        # the first lookup fails, succeeds on next check
        s.side_effect = [socket.error, MagicMock()]
        addrs = network._test_addrs(addrinfo, 80)
        assert len(addrs) == 1
        assert addrs[0] == addrinfo[2][4][0]

        # attempt to connect to resolved address with default timeout
        s.side_effect = socket.error
        addrs = network._test_addrs(addrinfo, 80)
        assert not len(addrs) == 0

        # nothing can connect, but we've eliminated duplicates
        s.side_effect = socket.error
        addrs = network._test_addrs(addrinfo, 80)
        assert len(addrs) == 5


def test_is_subnet():
    for subnet_data in (IPV4_SUBNETS, IPV6_SUBNETS):
        for item in subnet_data[True]:
            log.debug("Testing that %s is a valid subnet", item)
            assert network.is_subnet(item)
        for item in subnet_data[False]:
            log.debug("Testing that %s is not a valid subnet", item)
            assert not network.is_subnet(item)


def test_is_ipv4_subnet():
    for item in IPV4_SUBNETS[True]:
        log.debug("Testing that %s is a valid subnet", item)
        assert network.is_ipv4_subnet(item)
    for item in IPV4_SUBNETS[False]:
        log.debug("Testing that %s is not a valid subnet", item)
        assert not network.is_ipv4_subnet(item)


def test_is_ipv6_subnet():
    for item in IPV6_SUBNETS[True]:
        log.debug("Testing that %s is a valid subnet", item)
        assert network.is_ipv6_subnet(item) is True
    for item in IPV6_SUBNETS[False]:
        log.debug("Testing that %s is not a valid subnet", item)
        assert network.is_ipv6_subnet(item) is False


@pytest.mark.parametrize(
    "addr,expected",
    (
        (24, "255.255.255.0"),
        (21, "255.255.248.0"),
        (17, "255.255.128.0"),
        (9, "255.128.0.0"),
        (36, ""),
        ("lol", ""),
    ),
)
def test_cidr_to_ipv4_netmask(addr, expected):
    assert network.cidr_to_ipv4_netmask(addr) == expected


def test_number_of_set_bits_to_ipv4_netmask():
    set_bits_to_netmask = network._number_of_set_bits_to_ipv4_netmask(0xFFFFFF00)
    assert set_bits_to_netmask == "255.255.255.0"
    set_bits_to_netmask = network._number_of_set_bits_to_ipv4_netmask(0xFFFF6400)
    assert set_bits_to_netmask == "255.255.100.0"


@pytest.mark.parametrize(
    "hex_num,inversion,expected",
    (
        ("0x4A7D2B63", False, "74.125.43.99"),
        ("0x4A7D2B63", True, "99.43.125.74"),
        ("00000000000000000000FFFF7F000001", False, "127.0.0.1"),
        ("0000000000000000FFFF00000100007F", True, "127.0.0.1"),
        ("20010DB8000000000000000000000000", False, "2001:db8::"),
        ("B80D0120000000000000000000000000", True, "2001:db8::"),
    ),
)
def test_hex2ip(hex_num, inversion, expected):
    assert network.hex2ip(hex_num, inversion) == expected


def test_interfaces_ifconfig_linux(linux_interfaces_dict):
    interfaces = network._interfaces_ifconfig(LINUX)
    assert interfaces == linux_interfaces_dict


def test_interfaces_ifconfig_freebsd(freebsd_interfaces_dict):
    interfaces = network._interfaces_ifconfig(FREEBSD)
    assert interfaces == freebsd_interfaces_dict


def test_interfaces_ifconfig_solaris():
    with patch("salt.utils.platform.is_sunos", return_value=True):
        expected_interfaces = {
            "ilbint0": {
                "inet6": [],
                "inet": [
                    {
                        "broadcast": "10.6.0.255",
                        "netmask": "255.255.255.0",
                        "address": "10.6.0.11",
                    }
                ],
                "up": True,
            },
            "lo0": {
                "inet6": [{"prefixlen": "128", "address": "::1"}],
                "inet": [{"netmask": "255.0.0.0", "address": "127.0.0.1"}],
                "up": True,
            },
            "ilbext0": {
                "inet6": [],
                "inet": [
                    {
                        "broadcast": "10.10.11.31",
                        "netmask": "255.255.255.224",
                        "address": "10.10.11.11",
                    },
                    {
                        "broadcast": "10.10.11.31",
                        "netmask": "255.255.255.224",
                        "address": "10.10.11.12",
                    },
                ],
                "up": True,
            },
            "vpn0": {
                "inet6": [],
                "inet": [{"netmask": "255.0.0.0", "address": "10.6.0.14"}],
                "up": True,
            },
            "net0": {
                "inet6": [{"prefixlen": "10", "address": "fe80::221:9bff:fefd:2a22"}],
                "inet": [
                    {
                        "broadcast": "10.10.10.63",
                        "netmask": "255.255.255.224",
                        "address": "10.10.10.38",
                    }
                ],
                "up": True,
            },
        }
        interfaces = network._interfaces_ifconfig(SOLARIS)
        assert interfaces == expected_interfaces


def test_interfaces_ifconfig_netbsd():
    expected_interfaces = {
        "lo0": {
            "inet": [{"address": "127.0.0.1", "netmask": "255.0.0.0"}],
            "inet6": [{"address": "fe80::1", "prefixlen": "64", "scope": "lo0"}],
            "up": True,
        },
        "vioif0": {
            "hwaddr": "00:a0:98:e6:83:18",
            "inet": [
                {
                    "address": "192.168.1.80",
                    "broadcast": "192.168.1.255",
                    "netmask": "255.255.255.0",
                }
            ],
            "inet6": [
                {
                    "address": "fe80::2a0:98ff:fee6:8318",
                    "prefixlen": "64",
                    "scope": "vioif0",
                }
            ],
            "up": True,
        },
    }
    interfaces = network._netbsd_interfaces_ifconfig(NETBSD)
    assert interfaces == expected_interfaces


def test_freebsd_remotes_on():
    with patch("salt.utils.platform.is_sunos", return_value=False):
        with patch("salt.utils.platform.is_freebsd", return_value=True):
            with patch("subprocess.check_output", return_value=FREEBSD_SOCKSTAT):
                remotes = network._freebsd_remotes_on("4506", "remote")
                assert remotes == {"127.0.0.1"}


def test_freebsd_remotes_on_with_fat_pid():
    with patch("salt.utils.platform.is_sunos", return_value=False):
        with patch("salt.utils.platform.is_freebsd", return_value=True):
            with patch(
                "subprocess.check_output",
                return_value=FREEBSD_SOCKSTAT_WITH_FAT_PID,
            ):
                remotes = network._freebsd_remotes_on("4506", "remote")
                assert remotes == {"127.0.0.1"}


def test_netlink_tool_remote_on_a():
    with patch("salt.utils.platform.is_sunos", return_value=False):
        with patch("salt.utils.platform.is_linux", return_value=True):
            with patch("subprocess.check_output", return_value=LINUX_NETLINK_SS_OUTPUT):
                remotes = network._netlink_tool_remote_on("4506", "local_port")
                assert remotes == {"192.168.122.177", "127.0.0.1"}


def test_netlink_tool_remote_on_b():
    with patch("subprocess.check_output", return_value=LINUX_NETLINK_SS_OUTPUT):
        remotes = network._netlink_tool_remote_on("4505", "remote_port")
        assert remotes == {"127.0.0.1", "1.2.3.4"}


def test_openbsd_remotes_on():
    with patch("subprocess.check_output", return_value=OPENBSD_NETSTAT):
        remotes = network._openbsd_remotes_on("4506", "remote")
        assert remotes == {"127.0.0.1"}


def test_openbsd_remotes_on_issue_61966():
    """
    Test that the command output is correctly converted to string before
    treating it as such
    """
    with patch("subprocess.check_output", return_value=OPENBSD_NETSTAT.encode()):
        remotes = network._openbsd_remotes_on("4506", "remote")
        assert remotes == {"127.0.0.1"}


def test_generate_minion_id_distinct():
    """
    Test if minion IDs are distinct in the pool.
    """
    with patch("platform.node", MagicMock(return_value="nodename")), patch(
        "socket.gethostname", MagicMock(return_value="hostname")
    ), patch(
        "socket.getfqdn", MagicMock(return_value="hostname.domainname.blank")
    ), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "attrname", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["1.2.3.4", "5.6.7.8"]),
    ):
        assert network._generate_minion_id() == [
            "hostname.domainname.blank",
            "nodename",
            "hostname",
            "1.2.3.4",
            "5.6.7.8",
        ]


def test_generate_minion_id_127_name():
    """
    Test if minion IDs can be named 127.foo
    """
    with patch("platform.node", MagicMock(return_value="127")), patch(
        "socket.gethostname", MagicMock(return_value="127")
    ), patch("socket.getfqdn", MagicMock(return_value="127.domainname.blank")), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "attrname", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["1.2.3.4", "5.6.7.8"]),
    ):
        assert network._generate_minion_id() == [
            "127.domainname.blank",
            "127",
            "1.2.3.4",
            "5.6.7.8",
        ]


def test_generate_minion_id_127_name_startswith():
    """
    Test if minion IDs can be named starting from "127"
    """
    expected = [
        "127890.domainname.blank",
        "127890",
        "1.2.3.4",
        "5.6.7.8",
    ]
    with patch("platform.node", MagicMock(return_value="127890")), patch(
        "socket.gethostname", MagicMock(return_value="127890")
    ), patch(
        "socket.getfqdn", MagicMock(return_value="127890.domainname.blank")
    ), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "attrname", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["1.2.3.4", "5.6.7.8"]),
    ):
        assert network._generate_minion_id() == expected


def test_generate_minion_id_duplicate():
    """
    Test if IP addresses in the minion IDs are distinct in the pool
    """
    expected = ["hostname", "1.2.3.4"]
    with patch("platform.node", MagicMock(return_value="hostname")), patch(
        "socket.gethostname", MagicMock(return_value="hostname")
    ), patch("socket.getfqdn", MagicMock(return_value="hostname")), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "hostname", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["1.2.3.4", "1.2.3.4", "1.2.3.4"]),
    ):
        assert network._generate_minion_id() == expected


def test_generate_minion_id_platform_used():
    """
    Test if platform.node is used for the first occurrence.
    The platform.node is most common hostname resolver before anything else.
    """
    with patch(
        "platform.node", MagicMock(return_value="very.long.and.complex.domain.name")
    ), patch("socket.gethostname", MagicMock(return_value="hostname")), patch(
        "socket.getfqdn", MagicMock(return_value="")
    ), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "hostname", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["1.2.3.4", "1.2.3.4", "1.2.3.4"]),
    ):
        assert network.generate_minion_id() == "very.long.and.complex.domain.name"


def test_generate_minion_id_platform_localhost_filtered():
    """
    Test if localhost is filtered from the first occurrence.
    """
    with patch("platform.node", MagicMock(return_value="localhost")), patch(
        "socket.gethostname", MagicMock(return_value="pick.me")
    ), patch(
        "socket.getfqdn", MagicMock(return_value="hostname.domainname.blank")
    ), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "hostname", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["1.2.3.4", "1.2.3.4", "1.2.3.4"]),
    ):
        assert network.generate_minion_id() == "hostname.domainname.blank"


def test_generate_minion_id_platform_localhost_filtered_all():
    """
    Test if any of the localhost is filtered from everywhere.
    """
    with patch("platform.node", MagicMock(return_value="localhost")), patch(
        "socket.gethostname", MagicMock(return_value="ip6-loopback")
    ), patch("socket.getfqdn", MagicMock(return_value="ip6-localhost")), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "localhost", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["127.0.0.1", "::1", "fe00::0", "fe02::1", "1.2.3.4"]),
    ):
        assert network.generate_minion_id() == "1.2.3.4"


def test_generate_minion_id_platform_localhost_only():
    """
    Test if there is no other choice but localhost.
    """
    with patch("platform.node", MagicMock(return_value="localhost")), patch(
        "socket.gethostname", MagicMock(return_value="ip6-loopback")
    ), patch("socket.getfqdn", MagicMock(return_value="ip6-localhost")), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "localhost", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["127.0.0.1", "::1", "fe00::0", "fe02::1"]),
    ):
        assert network.generate_minion_id() == "localhost"


def test_generate_minion_id_platform_fqdn():
    """
    Test if fqdn is picked up.
    """
    with patch("platform.node", MagicMock(return_value="localhost")), patch(
        "socket.gethostname", MagicMock(return_value="ip6-loopback")
    ), patch("socket.getfqdn", MagicMock(return_value="pick.me")), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "localhost", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["127.0.0.1", "::1", "fe00::0", "fe02::1"]),
    ):
        assert network.generate_minion_id() == "pick.me"


def test_generate_minion_id_platform_localhost_addrinfo():
    """
    Test if addinfo is picked up.
    """
    with patch("platform.node", MagicMock(return_value="localhost")), patch(
        "socket.gethostname", MagicMock(return_value="ip6-loopback")
    ), patch("socket.getfqdn", MagicMock(return_value="ip6-localhost")), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "pick.me", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["127.0.0.1", "::1", "fe00::0", "fe02::1"]),
    ):
        assert network.generate_minion_id() == "pick.me"


def test_generate_minion_id_platform_ip_addr_only():
    """
    Test if IP address is the only what is used as a Minion ID in case no DNS name.
    """
    with patch("platform.node", MagicMock(return_value="localhost")), patch(
        "socket.gethostname", MagicMock(return_value="ip6-loopback")
    ), patch("socket.getfqdn", MagicMock(return_value="ip6-localhost")), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "localhost", ("127.0.1.1", 0))]),
    ), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch(
        "salt.utils.network.ip_addrs",
        MagicMock(return_value=["127.0.0.1", "::1", "fe00::0", "fe02::1", "1.2.3.4"]),
    ):
        assert network.generate_minion_id() == "1.2.3.4"


def test_gen_mac():
    expected_mac = "00:16:3E:01:01:01"
    with patch("random.randint", return_value=1) as random_mock:
        assert random_mock.return_value == 1
        ret = network.gen_mac("00:16:3E")
        assert ret == expected_mac


@pytest.mark.parametrize(
    "mac_addr",
    (
        ("31337"),
        ("0001020304056"),
        ("00:01:02:03:04:056"),
        ("a0:b0:c0:d0:e0:fg"),
    ),
)
def test_mac_str_to_bytes_exceptions(mac_addr):
    with pytest.raises(ValueError):
        network.mac_str_to_bytes(mac_addr)


def test_mac_str_to_bytes():
    assert network.mac_str_to_bytes("100806040200") == b"\x10\x08\x06\x04\x02\x00"
    assert network.mac_str_to_bytes("f8e7d6c5b4a3") == b"\xf8\xe7\xd6\xc5\xb4\xa3"


@pytest.mark.slow_test
def test_generate_minion_id_with_long_hostname():
    """
    Validate the fix for:

    https://github.com/saltstack/salt/issues/51160
    """
    long_name = "localhost-abcdefghijklmnopqrstuvwxyz-abcdefghijklmnopqrstuvwxyz"
    with patch("socket.gethostname", MagicMock(return_value=long_name)):
        # An exception is raised if unicode is passed to socket.getfqdn
        minion_id = network.generate_minion_id()
    assert minion_id != ""


def test_filter_by_networks_with_no_filter():
    ips = ["10.0.123.200", "10.10.10.10"]
    with pytest.raises(TypeError):
        network.filter_by_networks(ips)  # pylint: disable=no-value-for-parameter


def test_filter_by_networks_empty_filter():
    ips = ["10.0.123.200", "10.10.10.10"]
    assert network.filter_by_networks(ips, []) == []


def test_filter_by_networks_ips_list():
    ips = [
        "10.0.123.200",
        "10.10.10.10",
        "193.124.233.5",
        "fe80::d210:cf3f:64e7:5423",
    ]
    expected = [
        "10.0.123.200",
        "10.10.10.10",
        "fe80::d210:cf3f:64e7:5423",
    ]
    networks = ["10.0.0.0/8", "fe80::/64"]
    assert network.filter_by_networks(ips, networks) == expected


def test_filter_by_networks_interfaces_dict():
    interfaces = {
        "wlan0": ["192.168.1.100", "217.5.140.67", "2001:db8::ff00:42:8329"],
        "eth0": [
            "2001:0DB8:0:CD30:123:4567:89AB:CDEF",
            "192.168.1.101",
            "10.0.123.201",
        ],
    }
    expected = {
        "wlan0": ["192.168.1.100", "2001:db8::ff00:42:8329"],
        "eth0": ["2001:0DB8:0:CD30:123:4567:89AB:CDEF", "192.168.1.101"],
    }
    ret = network.filter_by_networks(interfaces, ["192.168.1.0/24", "2001:db8::/48"])
    assert ret == expected


def test_filter_by_networks_catch_all():
    ips = [
        "10.0.123.200",
        "10.10.10.10",
        "193.124.233.5",
        "fe80::d210:cf3f:64e7:5423",
    ]
    assert network.filter_by_networks(ips, ["0.0.0.0/0", "::/0"]) == ips


def test_ip_networks():
    # We don't need to test with each platform's ifconfig/iproute2 output,
    # since this test isn't testing getting the interfaces. We already have
    # tests for that.
    interface_data = network._interfaces_ifconfig(LINUX)

    # Without loopback
    ret = network.ip_networks(interface_data=interface_data)
    assert ret == ["10.10.8.0/22"]
    # Without loopback, specific interface
    ret = network.ip_networks(interface="eth0", interface_data=interface_data)
    assert ret == ["10.10.8.0/22"]
    # Without loopback, multiple specific interfaces
    ret = network.ip_networks(interface="eth0,lo", interface_data=interface_data)
    assert ret == ["10.10.8.0/22"]
    # Without loopback, specific interface (not present)
    ret = network.ip_networks(interface="eth1", interface_data=interface_data)
    assert ret == []
    # With loopback
    ret = network.ip_networks(include_loopback=True, interface_data=interface_data)
    assert ret == ["10.10.8.0/22", "127.0.0.0/8"]
    # With loopback, specific interface
    ret = network.ip_networks(
        interface="eth0", include_loopback=True, interface_data=interface_data
    )
    assert ret == ["10.10.8.0/22"]
    # With loopback, multiple specific interfaces
    ret = network.ip_networks(
        interface="eth0,lo", include_loopback=True, interface_data=interface_data
    )
    assert ret == ["10.10.8.0/22", "127.0.0.0/8"]
    # With loopback, specific interface (not present)
    ret = network.ip_networks(
        interface="eth1", include_loopback=True, interface_data=interface_data
    )
    assert ret == []

    # Verbose, without loopback
    ret = network.ip_networks(verbose=True, interface_data=interface_data)
    expected_ret1 = {
        "10.10.8.0/22": {
            "prefixlen": 22,
            "netmask": "255.255.252.0",
            "num_addresses": 1024,
            "address": "10.10.8.0",
        },
    }
    assert ret == expected_ret1

    # Verbose, without loopback, specific interface
    ret = network.ip_networks(
        interface="eth0", verbose=True, interface_data=interface_data
    )
    expected_ret2 = {
        "10.10.8.0/22": {
            "prefixlen": 22,
            "netmask": "255.255.252.0",
            "num_addresses": 1024,
            "address": "10.10.8.0",
        },
    }
    assert ret == expected_ret2

    # Verbose, without loopback, multiple specific interfaces
    ret = network.ip_networks(
        interface="eth0,lo", verbose=True, interface_data=interface_data
    )
    expected_ret3 = {
        "10.10.8.0/22": {
            "prefixlen": 22,
            "netmask": "255.255.252.0",
            "num_addresses": 1024,
            "address": "10.10.8.0",
        },
    }
    assert ret == expected_ret3

    # Verbose, without loopback, specific interface (not present)
    ret = network.ip_networks(
        interface="eth1", verbose=True, interface_data=interface_data
    )
    assert ret == {}
    # Verbose, with loopback
    ret = network.ip_networks(
        include_loopback=True, verbose=True, interface_data=interface_data
    )
    expected_ret4 = {
        "10.10.8.0/22": {
            "prefixlen": 22,
            "netmask": "255.255.252.0",
            "num_addresses": 1024,
            "address": "10.10.8.0",
        },
        "127.0.0.0/8": {
            "prefixlen": 8,
            "netmask": "255.0.0.0",
            "num_addresses": 16777216,
            "address": "127.0.0.0",
        },
    }
    assert ret == expected_ret4

    # Verbose, with loopback, specific interface
    ret = network.ip_networks(
        interface="eth0",
        include_loopback=True,
        verbose=True,
        interface_data=interface_data,
    )
    expected_ret5 = {
        "10.10.8.0/22": {
            "prefixlen": 22,
            "netmask": "255.255.252.0",
            "num_addresses": 1024,
            "address": "10.10.8.0",
        },
    }
    assert ret == expected_ret5

    # Verbose, with loopback, multiple specific interfaces
    ret = network.ip_networks(
        interface="eth0,lo",
        include_loopback=True,
        verbose=True,
        interface_data=interface_data,
    )
    expected_ret6 = {
        "10.10.8.0/22": {
            "prefixlen": 22,
            "netmask": "255.255.252.0",
            "num_addresses": 1024,
            "address": "10.10.8.0",
        },
        "127.0.0.0/8": {
            "prefixlen": 8,
            "netmask": "255.0.0.0",
            "num_addresses": 16777216,
            "address": "127.0.0.0",
        },
    }
    assert ret == expected_ret6

    # Verbose, with loopback, specific interface (not present)
    ret = network.ip_networks(
        interface="eth1",
        include_loopback=True,
        verbose=True,
        interface_data=interface_data,
    )
    assert ret == {}


def test_ip_networks6():
    # We don't need to test with each platform's ifconfig/iproute2 output,
    # since this test isn't testing getting the interfaces. We already have
    # tests for that.
    interface_data = network._interfaces_ifconfig(LINUX)

    # Without loopback
    ret = network.ip_networks6(interface_data=interface_data)
    assert ret == ["fe80::/64"]
    # Without loopback, specific interface
    ret = network.ip_networks6(interface="eth0", interface_data=interface_data)
    assert ret == ["fe80::/64"]
    # Without loopback, multiple specific interfaces
    ret = network.ip_networks6(interface="eth0,lo", interface_data=interface_data)
    assert ret == ["fe80::/64"]
    # Without loopback, specific interface (not present)
    ret = network.ip_networks6(interface="eth1", interface_data=interface_data)
    assert ret == []
    # With loopback
    ret = network.ip_networks6(include_loopback=True, interface_data=interface_data)
    assert ret == ["::1/128", "fe80::/64"]
    # With loopback, specific interface
    ret = network.ip_networks6(
        interface="eth0", include_loopback=True, interface_data=interface_data
    )
    assert ret == ["fe80::/64"]
    # With loopback, multiple specific interfaces
    ret = network.ip_networks6(
        interface="eth0,lo", include_loopback=True, interface_data=interface_data
    )
    assert ret == ["::1/128", "fe80::/64"]
    # With loopback, specific interface (not present)
    ret = network.ip_networks6(
        interface="eth1", include_loopback=True, interface_data=interface_data
    )
    assert ret == []

    # Verbose, without loopback
    ret = network.ip_networks6(verbose=True, interface_data=interface_data)
    expected_ret1 = {
        "fe80::/64": {
            "prefixlen": 64,
            "netmask": "ffff:ffff:ffff:ffff::",
            "num_addresses": 18446744073709551616,
            "address": "fe80::",
        },
    }
    assert ret == expected_ret1

    # Verbose, without loopback, specific interface
    ret = network.ip_networks6(
        interface="eth0", verbose=True, interface_data=interface_data
    )
    expected_ret2 = {
        "fe80::/64": {
            "prefixlen": 64,
            "netmask": "ffff:ffff:ffff:ffff::",
            "num_addresses": 18446744073709551616,
            "address": "fe80::",
        },
    }
    assert ret == expected_ret2

    # Verbose, without loopback, multiple specific interfaces
    ret = network.ip_networks6(
        interface="eth0,lo", verbose=True, interface_data=interface_data
    )
    expected_ret3 = {
        "fe80::/64": {
            "prefixlen": 64,
            "netmask": "ffff:ffff:ffff:ffff::",
            "num_addresses": 18446744073709551616,
            "address": "fe80::",
        },
    }
    assert ret == expected_ret3

    # Verbose, without loopback, specific interface (not present)
    ret = network.ip_networks6(
        interface="eth1", verbose=True, interface_data=interface_data
    )
    assert ret == {}

    # Verbose, with loopback
    ret = network.ip_networks6(
        include_loopback=True, verbose=True, interface_data=interface_data
    )
    expected_ret4 = {
        "fe80::/64": {
            "prefixlen": 64,
            "netmask": "ffff:ffff:ffff:ffff::",
            "num_addresses": 18446744073709551616,
            "address": "fe80::",
        },
        "::1/128": {
            "prefixlen": 128,
            "netmask": "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff",
            "num_addresses": 1,
            "address": "::1",
        },
    }
    assert ret == expected_ret4

    # Verbose, with loopback, specific interface
    ret = network.ip_networks6(
        interface="eth0",
        include_loopback=True,
        verbose=True,
        interface_data=interface_data,
    )
    expected_ret5 = {
        "fe80::/64": {
            "prefixlen": 64,
            "netmask": "ffff:ffff:ffff:ffff::",
            "num_addresses": 18446744073709551616,
            "address": "fe80::",
        },
    }
    assert ret == expected_ret5

    # Verbose, with loopback, multiple specific interfaces
    ret = network.ip_networks6(
        interface="eth0,lo",
        include_loopback=True,
        verbose=True,
        interface_data=interface_data,
    )
    expected_ret6 = {
        "fe80::/64": {
            "prefixlen": 64,
            "netmask": "ffff:ffff:ffff:ffff::",
            "num_addresses": 18446744073709551616,
            "address": "fe80::",
        },
        "::1/128": {
            "prefixlen": 128,
            "netmask": "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff",
            "num_addresses": 1,
            "address": "::1",
        },
    }
    assert ret == expected_ret6

    # Verbose, with loopback, specific interface (not present)
    ret = network.ip_networks6(
        interface="eth1",
        include_loopback=True,
        verbose=True,
        interface_data=interface_data,
    )
    assert ret == {}


def test_get_fqhostname_return():
    """
    Test if proper hostname is used when RevDNS differ from hostname
    """
    with patch("socket.gethostname", MagicMock(return_value="hostname")), patch(
        "socket.getfqdn",
        MagicMock(return_value="very.long.and.complex.domain.name"),
    ), patch(
        "socket.getaddrinfo",
        MagicMock(return_value=[(2, 3, 0, "hostname", ("127.0.1.1", 0))]),
    ):
        assert network.get_fqhostname() == "hostname"


def test_get_fqhostname_return_empty_hostname():
    """
    Test if proper hostname is used when hostname returns empty string
    """
    host = "hostname"
    with patch("socket.gethostname", MagicMock(return_value=host)), patch(
        "socket.getfqdn",
        MagicMock(return_value="very.long.and.complex.domain.name"),
    ), patch(
        "socket.getaddrinfo",
        MagicMock(
            return_value=[
                (2, 3, 0, host, ("127.0.1.1", 0)),
                (2, 3, 0, "", ("127.0.1.1", 0)),
            ]
        ),
    ):
        assert network.get_fqhostname() == host


@pytest.mark.parametrize(
    "addr,expected,strip",
    (
        ("127.0.0.1", "127.0.0.1", False),
        ("[::1]", "::1", True),
        ("::1", "[::1]", False),
        ("[::1]", "[::1]", False),
        (ipaddress.ip_address("127.0.0.1"), "127.0.0.1", False),
    ),
)
def test_ip_bracket(addr, expected, strip):
    assert network.ip_bracket(addr, strip=strip) == expected


def test_junos_ifconfig_output_parsing():
    ret = network._junos_interfaces_ifconfig("inet mtu 0 local=" + " " * 3456)
    assert ret == {"inet": {"up": False}}


def test_isportopen_false():
    ret = network.isportopen("127.0.0.1", "66000")
    assert ret is False


@pytest.fixture
def openport_22233():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("0.0.0.0", 22233))
    s.listen(5)
    yield
    s.close()


def test_isportopen(openport_22233):
    ret = network.isportopen("127.0.0.1", 22233)
    assert ret == 0


def test_get_socket():
    ret = network.get_socket("127.0.0.1")
    assert ret.family == socket.AF_INET
    assert ret.type == socket.SOCK_STREAM

    ret = network.get_socket("2001:a71::1")
    assert ret.family == socket.AF_INET6
    assert ret.type == socket.SOCK_STREAM


# @pytest.mark.skip_on_windows(reason="Do not run on Windows")
def test_ip_to_host(grains):
    if salt.utils.platform.is_windows():
        hostname = socket.gethostname()
    else:
        hostname = "localhost"

    ret = network.ip_to_host("127.0.0.1")
    if grains.get("oscodename") == "Photon":
        # Photon returns this for IPv4
        assert ret in ("ipv6-localhost", "localhost")
    else:
        assert ret == hostname

    ret = network.ip_to_host("2001:a71::1")
    assert ret is None

    ret = network.ip_to_host("::1")
    if grains["os"] == "Amazon":
        assert ret in ("localhost6", "localhost")
    elif grains["os_family"] == "Debian":
        if grains["osmajorrelease"] == 12:
            assert ret == hostname
        else:
            assert ret in ("ip6-localhost", "localhost")
    elif grains["os_family"] == "RedHat":
        if grains["oscodename"] == "Photon":
            assert ret in ("ipv6-localhost", "localhost")
        else:
            assert ret == hostname
    elif grains["os_family"] == "Arch":
        if grains.get("osmajorrelease", None) is None:
            # running doesn't have osmajorrelease grains
            assert ret == hostname
        else:
            assert ret in ("ip6-localhost", "localhost")
    else:
        assert ret == hostname


@pytest.mark.parametrize(
    "addr,fmtr,expected",
    (
        ("192.168.0.115", "prefixlen", "/24"),
        ("192.168.1.80", "prefixlen", "/24"),
        ("10.10.10.250", "prefixlen", "/8"),
        ("192.168.0.115", "netmask", "255.255.255.0"),
        ("192.168.1.80", "netmask", "255.255.255.0"),
        ("10.10.10.250", "netmask", "255.0.0.0"),
    ),
)
def test_natural_ipv4_netmask(addr, fmtr, expected):
    assert network.natural_ipv4_netmask(addr, fmt=fmtr) == expected


@pytest.mark.parametrize(
    "addr,expected",
    (
        ("127.0", "127.0.0.0"),
        ("192.168.3", "192.168.3.0"),
        ("10.209", "10.209.0.0"),
    ),
)
def test_rpad_ipv4_network(addr, expected):
    assert network.rpad_ipv4_network(addr) == expected


def test_hw_addr(linux_interfaces_dict, freebsd_interfaces_dict):

    with patch(
        "salt.utils.network.interfaces",
        MagicMock(return_value=linux_interfaces_dict),
    ):
        hw_addrs = network.hw_addr("eth0")
        assert hw_addrs == "e0:3f:49:85:6a:af"

    with patch(
        "salt.utils.network.interfaces", MagicMock(return_value=freebsd_interfaces_dict)
    ), patch("salt.utils.platform.is_netbsd", MagicMock(return_value=True)):
        hw_addrs = network.hw_addr("em0")
        assert hw_addrs == "00:30:48:ff:ff:ff"

        hw_addrs = network.hw_addr("em1")
        assert hw_addrs == "00:30:48:aa:aa:aa"

        hw_addrs = network.hw_addr("dog")
        assert (
            hw_addrs
            == 'Interface "dog" not in available interfaces: "", "em0", "em1", "lo0", "plip0", "tun0"'
        )


def test_interface_and_ip(linux_interfaces_dict):

    with patch(
        "salt.utils.network.interfaces",
        MagicMock(return_value=linux_interfaces_dict),
    ):
        expected = [
            {
                "address": "10.10.10.56",
                "broadcast": "10.10.10.255",
                "netmask": "255.255.252.0",
            }
        ]
        ret = network.interface("eth0")
        assert ret == expected

        ret = network.interface("dog")
        assert ret == 'Interface "dog" not in available interfaces: "eth0", "lo"'

        ret = network.interface_ip("eth0")
        assert ret == "10.10.10.56"

        ret = network.interface_ip("dog")
        assert ret == 'Interface "dog" not in available interfaces: "eth0", "lo"'


def test_subnets(linux_interfaces_dict):

    with patch(
        "salt.utils.network.interfaces",
        MagicMock(return_value=linux_interfaces_dict),
    ):
        ret = network.subnets()
        assert ret == ["10.10.8.0/22"]

        ret = network.subnets6()
        assert ret == ["fe80::/64"]


def test_in_subnet(caplog):
    assert network.in_subnet("fe80::/64", "fe80::e23f:49ff:fe85:6aaf")
    assert network.in_subnet("10.10.8.0/22", "10.10.10.56")
    assert not network.in_subnet("10.10.8.0/22")

    caplog.clear()
    ret = network.in_subnet("10.10.8.0/40")
    assert "Invalid CIDR '10.10.8.0/40'" in caplog.text
    assert not ret


def test_ip_addrs(linux_interfaces_dict):
    with patch(
        "salt.utils.network.interfaces",
        MagicMock(return_value=linux_interfaces_dict),
    ):
        ret = network.ip_addrs("eth0")
        assert ret == ["10.10.10.56"]

    with patch(
        "salt.utils.network.interfaces",
        MagicMock(return_value=linux_interfaces_dict),
    ):
        ret = network.ip_addrs6("eth0")
        assert ret == ["fe80::e23f:49ff:fe85:6aaf"]



import collections
import textwrap

import pytest

import salt.roster.sshconfig as sshconfig
from tests.support.mock import mock_open, patch


@pytest.fixture
def target_abc():
    return collections.OrderedDict(
        [
            ("user", "user.mcuserface"),
            ("priv", "~/.ssh/id_rsa_abc"),
            ("host", "abc.asdfgfdhgjkl.com"),
        ]
    )


@pytest.fixture
def target_abc123():
    return collections.OrderedDict(
        [
            ("user", "user.mcuserface"),
            ("priv", "~/.ssh/id_rsa_abc"),
            ("host", "abc123.asdfgfdhgjkl.com"),
        ]
    )


@pytest.fixture
def target_def():
    return collections.OrderedDict(
        [
            ("user", "user.mcuserface"),
            ("priv", "~/.ssh/id_rsa_def"),
            ("host", "def.asdfgfdhgjkl.com"),
        ]
    )


@pytest.fixture
def all_(target_abc, target_abc123, target_def):
    return {
        "abc.asdfgfdhgjkl.com": target_abc,
        "abc123.asdfgfdhgjkl.com": target_abc123,
        "def.asdfgfdhgjkl.com": target_def,
    }


@pytest.fixture
def abc_glob(target_abc, target_abc123):
    return {
        "abc.asdfgfdhgjkl.com": target_abc,
        "abc123.asdfgfdhgjkl.com": target_abc123,
    }


@pytest.fixture
def mock_fp():
    sample_ssh_config = textwrap.dedent(
        """
    Host *
        User user.mcuserface

    Host abc*
        IdentityFile ~/.ssh/id_rsa_abc

    Host def*
        IdentityFile  ~/.ssh/id_rsa_def

    Host abc.asdfgfdhgjkl.com
        HostName 123.123.123.123

    Host abc123.asdfgfdhgjkl.com
        HostName 123.123.123.124

    Host def.asdfgfdhgjkl.com
        HostName      234.234.234.234
    """
    )

    return mock_open(read_data=sample_ssh_config)


@pytest.fixture
def configure_loader_modules():
    return {sshconfig: {}}


def test_all(mock_fp, all_):
    with patch("salt.utils.files.fopen", mock_fp):
        with patch("salt.roster.sshconfig._get_ssh_config_file"):
            targets = sshconfig.targets("*")
    assert targets == all_


def test_abc_glob(mock_fp, abc_glob):
    with patch("salt.utils.files.fopen", mock_fp):
        with patch("salt.roster.sshconfig._get_ssh_config_file"):
            targets = sshconfig.targets("abc*")
    assert targets == abc_glob



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.ilo
"""

import tempfile

import pytest

import salt.modules.file
import salt.modules.ilo as ilo
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        ilo: {
            "__opts__": {"cachedir": tempfile.gettempdir()},
            "__salt__": {"file.remove": salt.modules.file.remove},
        }
    }


# '__execute_cmd' function tests: 1


def test_execute_cmd():
    """
    Test if __execute_command opens the temporary file
    properly when calling global_settings.
    """
    mock_cmd_run = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(ilo.__salt__, {"cmd.run_all": mock_cmd_run}):
        ret = ilo.global_settings()
        assert ret  # == True


# 'global_settings' function tests: 1


def test_global_settings():
    """
    Test if it shows global_settings
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Global Settings": {}}),
    ):
        assert ilo.global_settings() == {"Global Settings": {}}


# 'set_http_port' function tests: 1


def test_set_http_port():
    """
    Test if it configure the port HTTP should listen on
    """
    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"HTTP_PORT": {"VALUE": 80}}},
    ):
        assert ilo.set_http_port()

    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"HTTP_PORT": {"VALUE": 40}}},
    ):
        with patch.object(ilo, "__execute_cmd", return_value={"Set HTTP Port": {}}):
            assert ilo.set_http_port() == {"Set HTTP Port": {}}


# 'set_https_port' function tests: 1


def test_set_https_port():
    """
    Test if it configure the port HTTPS should listen on
    """
    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"HTTP_PORT": {"VALUE": 443}}},
    ):
        assert ilo.set_https_port()

    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"HTTP_PORT": {"VALUE": 80}}},
    ):
        with patch.object(ilo, "__execute_cmd", return_value={"Set HTTPS Port": {}}):
            assert ilo.set_https_port() == {"Set HTTPS Port": {}}


# 'enable_ssh' function tests: 1


def test_enable_ssh():
    """
    Test if it enable the SSH daemon
    """
    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"SSH_STATUS": {"VALUE": "Y"}}},
    ):
        assert ilo.enable_ssh()

    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"SSH_STATUS": {"VALUE": "N"}}},
    ):
        with patch.object(ilo, "__execute_cmd", return_value={"Enable SSH": {}}):
            assert ilo.enable_ssh() == {"Enable SSH": {}}


# 'disable_ssh' function tests: 1


def test_disable_ssh():
    """
    Test if it disable the SSH daemon
    """
    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"SSH_STATUS": {"VALUE": "N"}}},
    ):
        assert ilo.disable_ssh()

    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"SSH_STATUS": {"VALUE": "Y"}}},
    ):
        with patch.object(ilo, "__execute_cmd", return_value={"Disable SSH": {}}):
            assert ilo.disable_ssh() == {"Disable SSH": {}}


# 'set_ssh_port' function tests: 1


def test_set_ssh_port():
    """
    Test if it enable SSH on a user defined port
    """
    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"SSH_PORT": {"VALUE": 22}}},
    ):
        assert ilo.set_ssh_port()

    with patch.object(
        ilo,
        "global_settings",
        return_value={"Global Settings": {"SSH_PORT": {"VALUE": 20}}},
    ):
        with patch.object(
            ilo, "__execute_cmd", return_value={"Configure SSH Port": {}}
        ):
            assert ilo.set_ssh_port() == {"Configure SSH Port": {}}


# 'set_ssh_key' function tests: 1


def test_set_ssh_key():
    """
    Test if it configure SSH public keys for specific users
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Import SSH Publickey": {}}),
    ):
        assert ilo.set_ssh_key("ssh-rsa AAAAB3Nza Salt") == {"Import SSH Publickey": {}}


# 'delete_ssh_key' function tests: 1


def test_delete_ssh_key():
    """
    Test if it delete a users SSH key from the ILO
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Delete user SSH key": {}}),
    ):
        assert ilo.delete_ssh_key("Salt") == {"Delete user SSH key": {}}


# 'list_users' function tests: 1


def test_list_users():
    """
    Test if it list all users
    """
    with patch(
        "salt.modules.ilo.__execute_cmd", MagicMock(return_value={"All users": {}})
    ):
        assert ilo.list_users() == {"All users": {}}


# 'list_users_info' function tests: 1


def test_list_users_info():
    """
    Test if it List all users in detail
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"All users info": {}}),
    ):
        assert ilo.list_users_info() == {"All users info": {}}


# 'create_user' function tests: 1


def test_create_user():
    """
    Test if it create user
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Create user": {}}),
    ):
        assert ilo.create_user("Salt", "secretagent", "VIRTUAL_MEDIA_PRIV") == {
            "Create user": {}
        }


# 'delete_user' function tests: 1


def test_delete_user():
    """
    Test if it delete a user
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Delete user": {}}),
    ):
        assert ilo.delete_user("Salt") == {"Delete user": {}}


# 'get_user' function tests: 1


def test_get_user():
    """
    Test if it returns local user information, excluding the password
    """
    with patch(
        "salt.modules.ilo.__execute_cmd", MagicMock(return_value={"User Info": {}})
    ):
        assert ilo.get_user("Salt") == {"User Info": {}}


# 'change_username' function tests: 1


def test_change_username():
    """
    Test if it change a username
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Change username": {}}),
    ):
        assert ilo.change_username("Salt", "SALT") == {"Change username": {}}


# 'change_password' function tests: 1


def test_change_password():
    """
    Test if it reset a users password
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Change password": {}}),
    ):
        assert ilo.change_password("Salt", "saltpasswd") == {"Change password": {}}


# 'network' function tests: 1


def test_network():
    """
    Test if it grab the current network settings
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Network Settings": {}}),
    ):
        assert ilo.network() == {"Network Settings": {}}


# 'configure_network' function tests: 1


def test_configure_network():
    """
    Test if it configure Network Interface
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Configure_Network": {}}),
    ):
        ret = {
            "Network Settings": {
                "IP_ADDRESS": {"VALUE": "10.0.0.10"},
                "SUBNET_MASK": {"VALUE": "255.255.255.0"},
                "GATEWAY_IP_ADDRESS": {"VALUE": "10.0.0.1"},
            }
        }
        with patch.object(ilo, "network", return_value=ret):
            assert ilo.configure_network("10.0.0.10", "255.255.255.0", "10.0.0.1")

        with patch.object(ilo, "network", return_value=ret):
            with patch.object(
                ilo, "__execute_cmd", return_value={"Network Settings": {}}
            ):
                assert ilo.configure_network(
                    "10.0.0.100", "255.255.255.10", "10.0.0.10"
                ) == {"Network Settings": {}}


# 'enable_dhcp' function tests: 1


def test_enable_dhcp():
    """
    Test if it enable DHCP
    """
    with patch.object(
        ilo,
        "network",
        return_value={"Network Settings": {"DHCP_ENABLE": {"VALUE": "Y"}}},
    ):
        assert ilo.enable_dhcp()

    with patch.object(
        ilo,
        "network",
        return_value={"Network Settings": {"DHCP_ENABLE": {"VALUE": "N"}}},
    ):
        with patch.object(ilo, "__execute_cmd", return_value={"Enable DHCP": {}}):
            assert ilo.enable_dhcp() == {"Enable DHCP": {}}


# 'disable_dhcp' function tests: 1


def test_disable_dhcp():
    """
    Test if it disable DHCP
    """
    with patch.object(
        ilo,
        "network",
        return_value={"Network Settings": {"DHCP_ENABLE": {"VALUE": "N"}}},
    ):
        assert ilo.disable_dhcp()

    with patch.object(
        ilo,
        "network",
        return_value={"Network Settings": {"DHCP_ENABLE": {"VALUE": "Y"}}},
    ):
        with patch.object(ilo, "__execute_cmd", return_value={"Disable DHCP": {}}):
            assert ilo.disable_dhcp() == {"Disable DHCP": {}}


# 'configure_snmp' function tests: 1


def test_configure_snmp():
    """
    Test if it configure SNMP
    """
    with patch(
        "salt.modules.ilo.__execute_cmd",
        MagicMock(return_value={"Configure SNMP": {}}),
    ):
        assert ilo.configure_snmp("Salt") == {"Configure SNMP": {}}



import textwrap

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_issue_8343_accumulated_require_in(modules, tmp_path, state_tree):
    name = tmp_path / "testfile"
    sls_contents = f"""
    {name}:
      file.managed:
        - contents: |
                    #

    prepend-foo-accumulator-from-pillar:
      file.accumulated:
        - require_in:
          - file: prepend-foo-management
        - filename: {name}
        - text: |
                foo

    append-foo-accumulator-from-pillar:
      file.accumulated:
        - require_in:
          - file: append-foo-management
        - filename: {name}
        - text: |
                bar

    prepend-foo-management:
      file.blockreplace:
        - name: {name}
        - marker_start: "#-- start salt managed zonestart -- PLEASE, DO NOT EDIT"
        - marker_end: "#-- end salt managed zonestart --"
        - content: ''
        - prepend_if_not_found: True
        - backup: '.bak'
        - show_changes: True

    append-foo-management:
      file.blockreplace:
        - name: {name}
        - marker_start: "#-- start salt managed zoneend -- PLEASE, DO NOT EDIT"
        - marker_end: "#-- end salt managed zoneend --"
        - content: ''
        - append_if_not_found: True
        - backup: '.bak2'
        - show_changes: True
    """
    with pytest.helpers.temp_file(
        "issue-8343.sls", directory=state_tree, contents=sls_contents
    ):
        ret = modules.state.sls("issue-8343")
        for state_run in ret:
            assert state_run.result is True

    expected = textwrap.dedent(
        """\
    #-- start salt managed zonestart -- PLEASE, DO NOT EDIT
    foo
    #-- end salt managed zonestart --
    #
    #-- start salt managed zoneend -- PLEASE, DO NOT EDIT
    bar
    #-- end salt managed zoneend --
    """
    )

    assert name.read_text() == expected


def test_issue_11003_immutable_lazy_proxy_sum(modules, tmp_path, state_tree):
    # causes the Import-Module ServerManager error on Windows
    name = tmp_path / "testfile"
    sls_contents = f"""
    a{name}:
      file.absent:
        - name: {name}

    {name}:
      file.managed:
        - contents: |
                    #

    test-acc1:
      file.accumulated:
        - require_in:
          - file: final
        - filename: {name}
        - text: |
                bar

    test-acc2:
      file.accumulated:
        - watch_in:
          - file: final
        - filename: {name}
        - text: |
                baz

    final:
      file.blockreplace:
        - name: {name}
        - marker_start: "#-- start managed zone PLEASE, DO NOT EDIT"
        - marker_end: "#-- end managed zone"
        - content: ''
        - append_if_not_found: True
        - show_changes: True
    """
    with pytest.helpers.temp_file(
        "issue-11003.sls", directory=state_tree, contents=sls_contents
    ):
        ret = modules.state.sls("issue-11003")
        for state_run in ret:
            assert state_run.result is True

    contents = name.read_text().splitlines()
    begin = contents.index("#-- start managed zone PLEASE, DO NOT EDIT") + 1
    end = contents.index("#-- end managed zone")
    block_contents = contents[begin:end]
    for item in ("", "bar", "baz"):
        block_contents.remove(item)
    assert block_contents == []


def test_issue_60426(modules, tmp_path, state_tree):
    target_path = tmp_path.joinpath("etc/foo/bar")
    jinja_name = "foo_bar"
    jinja_contents = (
        "{% for item in accumulator['accumulated configstuff'] %}{{ item }}{% endfor %}"
    )

    sls_name = "issue-60426"
    sls_contents = f"""
    configuration file:
      file.managed:
        - name: {target_path}
        - source: salt://foo_bar.jinja
        - template: jinja
        - makedirs: True

    accumulated configstuff:
      file.accumulated:
        - filename: {target_path}
        - text:
          - some
          - good
          - stuff
        - watch_in:
          - configuration file
    """

    sls_tempfile = pytest.helpers.temp_file(
        f"{sls_name}.sls", directory=state_tree, contents=sls_contents
    )

    jinja_tempfile = pytest.helpers.temp_file(
        f"{jinja_name}.jinja", directory=state_tree, contents=jinja_contents
    )

    with sls_tempfile, jinja_tempfile:
        ret = modules.state.apply(sls_name)
        for state_run in ret:
            assert state_run.result is True
        # Check to make sure the file was created
        assert target_path.is_file()
        # The type of new line, ie, `\n` vs `\r\n` is not important
        assert target_path.read_text() == "somegoodstuff"

    sls_contents = f"""
    configuration file:
      file.managed:
        - name: {target_path}
        - source: salt://foo_bar.jinja
        - template: jinja
        - makedirs: True

    accumulated configstuff:
      file.accumulated:
        - filename: {target_path}
        - text:
          - some
          - more
          - good
          - stuff
        - watch_in:
          - file: configuration file
    """

    sls_tempfile = pytest.helpers.temp_file(
        f"{sls_name}.sls", directory=state_tree, contents=sls_contents
    )
    jinja_tempfile = pytest.helpers.temp_file(
        f"{jinja_name}.jinja", directory=state_tree, contents=jinja_contents
    )

    with sls_tempfile, jinja_tempfile:
        ret = modules.state.apply(sls_name)
        for state_run in ret:
            assert state_run.result is True
        # Check to make sure the file was created
        assert target_path.is_file()
        # The type of new line, ie, `\n` vs `\r\n` is not important
        assert target_path.read_text() == "somemoregoodstuff"



from collections import OrderedDict

import pytest

import salt.modules.pillar as pillarmod
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def pillar_value():
    pillar_value = dict(a=1, b="very secret")
    return pillar_value


@pytest.fixture
def configure_loader_modules():
    return {pillarmod: {}}


def test_obfuscate_inner_recursion():
    assert pillarmod._obfuscate_inner(
        dict(a=[1, 2], b=dict(pwd="secret", deeper=("a", 1)))
    ) == dict(a=["<int>", "<int>"], b=dict(pwd="<str>", deeper=("<str>", "<int>")))


def test_obfuscate_inner_more_types():
    assert pillarmod._obfuscate_inner(OrderedDict([("key", "value")])) == OrderedDict(
        [("key", "<str>")]
    )

    assert pillarmod._obfuscate_inner({1, 2}) == {"<int>"}

    assert pillarmod._obfuscate_inner((1, 2)) == ("<int>", "<int>")


def test_obfuscate(pillar_value):
    with patch("salt.modules.pillar.items", MagicMock(return_value=pillar_value)):
        assert pillarmod.obfuscate() == dict(a="<int>", b="<str>")


def test_obfuscate_with_kwargs(pillar_value):
    with patch(
        "salt.modules.pillar.items", MagicMock(return_value=pillar_value)
    ) as pillar_items_mock:
        ret = pillarmod.obfuscate(saltenv="saltenv")
        # ensure the kwargs are passed along to pillar.items
        assert call(saltenv="saltenv") in pillar_items_mock.mock_calls
        assert ret == dict(a="<int>", b="<str>")


def test_ls(pillar_value):
    with patch("salt.modules.pillar.items", MagicMock(return_value=pillar_value)):
        ls = sorted(pillarmod.ls())
        assert ls == ["a", "b"]


def test_pillar_get_default_merge():
    defaults = {
        "int": 1,
        "string": "foo",
        "list": ["foo"],
        "dict": {"foo": "bar", "subkey": {"foo": "bar"}},
    }

    pillar_mock = {
        "int": 2,
        "string": "bar",
        "list": ["bar", "baz"],
        "dict": {"baz": "qux", "subkey": {"baz": "qux"}},
    }

    # Test that we raise a KeyError when pillar_raise_on_missing is True
    with patch.dict(pillarmod.__opts__, {"pillar_raise_on_missing": True}):
        pytest.raises(KeyError, pillarmod.get, "missing")
    # Test that we return an empty string when it is not
    with patch.dict(pillarmod.__opts__, {}):
        assert pillarmod.get("missing") == ""

    with patch.dict(pillarmod.__pillar__, pillar_mock):
        # Test with no default passed (it should be KeyError) and merge=True.
        # The merge should be skipped and the value returned from __pillar__
        # should be returned.
        for item in pillarmod.__pillar__:
            assert pillarmod.get(item, merge=True) == pillarmod.__pillar__[item]

        # Test merging when the type of the default value is not the same as
        # what was returned. Merging should be skipped and the value returned
        # from __pillar__ should be returned.
        for default_type in defaults:
            for data_type in ("dict", "list"):
                if default_type == data_type:
                    continue
                assert (
                    pillarmod.get(data_type, default=defaults[default_type], merge=True)
                    == pillarmod.__pillar__[data_type]
                )

        # Test recursive dict merging
        assert pillarmod.get("dict", default=defaults["dict"], merge=True) == {
            "foo": "bar",
            "baz": "qux",
            "subkey": {"foo": "bar", "baz": "qux"},
        }

        # Test list merging
        assert pillarmod.get("list", default=defaults["list"], merge=True) == [
            "foo",
            "bar",
            "baz",
        ]


def test_pillar_get_default_merge_regression_38558():
    """Test for pillar.get(key=..., default=..., merge=True)
    Do not update the ``default`` value when using ``merge=True``.
    See: https://github.com/saltstack/salt/issues/38558
    """
    with patch.dict(pillarmod.__pillar__, {"l1": {"l2": {"l3": 42}}}):

        res = pillarmod.get(key="l1")
        assert {"l2": {"l3": 42}} == res

        default = {"l2": {"l3": 43}}

        res = pillarmod.get(key="l1", default=default)
        assert {"l2": {"l3": 42}} == res
        assert {"l2": {"l3": 43}} == default

        res = pillarmod.get(key="l1", default=default, merge=True)
        assert {"l2": {"l3": 42}} == res
        assert {"l2": {"l3": 43}} == default


def test_pillar_get_default_merge_regression_39062():
    """
    Confirm that we do not raise an exception if default is None and
    merge=True.
    See https://github.com/saltstack/salt/issues/39062 for more info.
    """
    with patch.dict(pillarmod.__pillar__, {"foo": "bar"}):

        assert pillarmod.get(key="foo", default=None, merge=True) == "bar"


def test_pillar_get_int_key():
    """
    Confirm that we can access pillar keys that are integers
    """
    with patch.dict(pillarmod.__pillar__, {12345: "luggage_code"}):

        assert pillarmod.get(key=12345, default=None, merge=True) == "luggage_code"

    with patch.dict(pillarmod.__pillar__, {12345: {"l2": {"l3": "my_luggage_code"}}}):

        res = pillarmod.get(key=12345)
        assert {"l2": {"l3": "my_luggage_code"}} == res

        default = {"l2": {"l3": "your_luggage_code"}}

        res = pillarmod.get(key=12345, default=default)
        assert {"l2": {"l3": "my_luggage_code"}} == res
        assert {"l2": {"l3": "your_luggage_code"}} == default

        res = pillarmod.get(key=12345, default=default, merge=True)
        assert {"l2": {"l3": "my_luggage_code"}} == res
        assert {"l2": {"l3": "your_luggage_code"}} == default


def test_pillar_keys():
    """
    Confirm that we can access pillar keys
    """
    with patch.dict(pillarmod.__pillar__, {"pkg": {"apache": "httpd"}}):
        test_key = "pkg"
        assert pillarmod.keys(test_key) == ["apache"]

    with patch.dict(
        pillarmod.__pillar__,
        {"12345": {"xyz": "my_luggage_code"}, "7": {"11": {"12": "13"}}},
    ):
        test_key = "7:11"
        res = pillarmod.keys(test_key)
        assert res == ["12"]



import logging

import pytest

import salt.returners.etcd_return as etcd_return
import salt.utils.json
from salt.utils.etcd_util import get_conn
from tests.support.pytest.etcd import *  # pylint: disable=wildcard-import,unused-wildcard-import

docker = pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_if_binaries_missing("docker", "dockerd", check_all=False),
]


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        etcd_return: {
            "__opts__": minion_opts,
        },
    }


@pytest.fixture(scope="module")
def update_etcd_profile(profile_name, prefix, etcd_profile):
    etcd_profile.update(
        {
            "etcd.returner": profile_name,
            "etcd.returner_root": prefix,
        }
    )

    return etcd_profile


@pytest.fixture(scope="module")
def minion_config_overrides(update_etcd_profile):
    return update_etcd_profile


@pytest.fixture(scope="module")
def etcd_client(minion_opts, profile_name):
    return get_conn(minion_opts, profile=profile_name)


@pytest.fixture(scope="module")
def prefix():
    return "/salt/pillar/test"


@pytest.fixture(autouse=True)
def cleanup_prefixed_entries(etcd_client, prefix):
    """
    Cleanup after each test to ensure a consistent starting state.
    """
    try:
        assert etcd_client.get(prefix, recurse=True) is None
        yield
    finally:
        etcd_client.delete(prefix, recurse=True)


def test_returner(prefix, etcd_client):
    """
    Test returning values to etcd
    """
    ret = {
        "id": "test-id",
        "jid": "123456789",
        "single-key": "single-value",
        "dict-key": {
            "dict-subkey-1": "subvalue-1",
            "dict-subkey-2": "subvalue-2",
        },
    }
    etcd_return.returner(ret)
    assert etcd_client.get("/".join((prefix, "minions", ret["id"]))) == ret["jid"]
    expected = {key: salt.utils.json.dumps(ret[key]) for key in ret}
    assert (
        etcd_client.get("/".join((prefix, "jobs", ret["jid"], ret["id"])), recurse=True)
        == expected
    )


def test_save_and_get_load():
    """
    Test saving a data load to etcd
    """
    jid = "123456789"
    load = {
        "single-key": "single-value",
        "dict-key": {
            "dict-subkey-1": "subvalue-1",
            "dict-subkey-2": "subvalue-2",
        },
    }
    etcd_return.save_load(jid, load)
    assert etcd_return.get_load(jid) == load


def test_get_jid():
    """
    Test getting the return for a given jid
    """
    jid = "123456789"
    ret = {
        "id": "test-id-1",
        "jid": jid,
        "single-key": "single-value",
        "dict-key": {
            "dict-subkey-1": "subvalue-1",
            "dict-subkey-2": "subvalue-2",
        },
        "return": "test-return-1",
    }
    etcd_return.returner(ret)

    ret = {"id": "test-id-2", "jid": jid, "return": "test-return-2"}
    etcd_return.returner(ret)

    expected = {
        "test-id-1": {"return": "test-return-1"},
        "test-id-2": {"return": "test-return-2"},
    }
    assert etcd_return.get_jid(jid) == expected


def test_get_fun():
    """
    Test getting the latest fn run for each minion and matching to a target fn
    """
    ret = {
        "id": "test-id-1",
        "jid": "1",
        "single-key": "single-value",
        "dict-key": {
            "dict-subkey-1": "subvalue-1",
            "dict-subkey-2": "subvalue-2",
        },
        "return": "test-return-1",
        "fun": "test.ping",
    }
    etcd_return.returner(ret)

    ret = {
        "id": "test-id-2",
        "jid": "2",
        "return": "test-return-2",
        "fun": "test.collatz",
    }
    etcd_return.returner(ret)

    expected = {
        "test-id-2": "test.collatz",
    }
    assert etcd_return.get_fun("test.collatz") == expected


def test_get_jids():
    """
    Test getting all jids
    """
    ret = {
        "id": "test-id-1",
        "jid": "1",
    }
    etcd_return.returner(ret)

    ret = {
        "id": "test-id-2",
        "jid": "2",
    }
    etcd_return.returner(ret)

    retval = etcd_return.get_jids()
    assert len(retval) == 2
    assert "1" in retval
    assert "2" in retval


def test_get_minions():
    """
    Test getting a list of minions
    """
    ret = {
        "id": "test-id-1",
        "jid": "1",
    }
    etcd_return.returner(ret)

    ret = {
        "id": "test-id-2",
        "jid": "2",
    }
    etcd_return.returner(ret)

    retval = etcd_return.get_minions()
    assert len(retval) == 2
    assert "test-id-1" in retval
    assert "test-id-2" in retval



"""
hjson renderer for Salt

See the hjson_ documentation for more information

.. _hjson: http://laktak.github.io/hjson/
"""

try:
    import hjson

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False


def render(hjson_data, saltenv="base", sls="", **kws):
    """
    Accepts HJSON as a string or as a file object and runs it through the HJSON
    parser.

    :rtype: A Python data structure
    """
    if not isinstance(hjson_data, str):
        hjson_data = hjson_data.read()

    if hjson_data.startswith("#!"):
        hjson_data = hjson_data[(hjson_data.find("\n") + 1) :]
    if not hjson_data.strip():
        return {}
    return hjson.loads(hjson_data)
